/*
   BundleManager.java

   Copyright (C) 2000 by ted stockwell
   All Rights Reserved.
 */
package org.servicetango.core;

import org.servicetango.util.*;
import org.servicetango.util.java.*;
import org.servicetango.log.*;
import org.osgi.service.log.*;
import org.osgi.framework.*;

import java.text.*;
import java.io.*;
import java.util.*;
import java.util.jar.*;
import java.beans.*;

import org.w3c.dom.*;
import org.xml.sax.*;

/**
 *  Encapsulates logic to manage bundles and create bundle context
 *  objects.
 *
 *  If a SecurityManager has not already been installed by the time a
 *  BundleManager is instantiated then an STSecurityManager will be used.
 *
 *  @author   ted stockwell [emorning@mediaone.net]
 */
public class BundleManager
{
  public static final int BEGIN = 1;
  public static final int STARTING = 2;
  public static final int STOPPING = 3;
  public static final int STARTED = 4;
  public static final int STOPPED = 5;

    Map m_bundleIDs= new HashMap();
    Map m_bundleNames= new HashMap();
    int m_state= BEGIN;

    static final Comparator BUNDLE_COMPARATOR=  new Comparator() {
        public int compare(Object o1, Object o2) {
            long id1= ((STBundle)o1).getBundleId();
            long id2= ((STBundle)o2).getBundleId();
            if (id1 < id2)
                return -1;
           if (id1 > id2)
            return 1;
            return 0;
        }
        public boolean equals(Object obj) {
            return obj == this;
        }
    };

    /**
     * A list of known bundles.
     * Some of these bundles may be uninstalled.
     */
    protected Set m_bundles= new HashSet();

    /**
     * An object that manages the framework's configuration
     */
    protected STConfiguration m_configuration= null;

    /**
     * An object that's used by this manager for logging messages
     */
    LogService m_log= new Log(new OutputStreamLogWriter(System.out));

    /**
     * An object that manages persistent storage for this bundle manager
     */
    STStorage m_storage;

    /**
     * A list of the ServiceRegistration objects associated with registered
     * services.
     */
    List m_registered= new ArrayList();

    /** Lists of listeners */
    protected EventListenerList m_listeners= new EventListenerList();

    static class ListenerEntry extends EventListenerList.Entry {
        STBundle m_bundle;
        String m_filter;
        ListenerEntry(Class c, EventListener l, STBundle b, String filter) {
            super(c, l);
            m_bundle= b;
            m_filter= filter;
        }
        STBundle getBundle() { return m_bundle; }
        String getFilter() { return m_filter; }
    }

  /**
   *    Create a BundleManager with the given properties.
   *
   *  If a SecurityManager has not already been installed by the time a
   *  BundleManager is instantiated then an STSecurityManager will be used.
   */
  public BundleManager(STConfiguration initProperties)
  throws BundleException
  {
        this(initProperties, null);
  }

  /**
   *    Create a BundleManager with the given properties.
   *    This manager will use...
   *    ...the given storage object to persist bundles
   *
   *  If a SecurityManager has not already been installed by the time a
   *  BundleManager is instantiated then an STSecurityManager will be used.
   */
  public BundleManager(STConfiguration initProperties, STStorage storage)
  throws BundleException
  {
    m_configuration= initProperties;

    m_storage= storage;
    if (m_storage == null)
        m_storage= new STStorage(this);

    init(initProperties);
  }

  protected void init(STConfiguration configuration) {
    SecurityManager sm= System.getSecurityManager();
    if (sm == null) {
        sm= new STSecurityManager();
        System.setSecurityManager(sm);
   }
  }


    /**
     *     Returns a reference to the context's Properties object
     */
    protected STConfiguration getConfiguration() {
      return m_configuration;
    }
    protected STStorage getStorage() {
        return m_storage;
    }
    protected LogService getLogService() {
        return m_log;
    }



    /**
    * starts bundles at startup.
    */
    protected void startUp()
    throws BundleException
    {
        setState(STARTING);
        try {
            List bundles;
            try {
                bundles= new ArrayList(m_storage.getInstalledBundles());
            }
            catch (Exception x) {
                throw new STException("Error while starting", x);
            }

            /*
            Sort the bundles according to thier bundle id.  This should
            preserve the order in which bundles were installed and started.
            */
            Collections.sort(bundles, BUNDLE_COMPARATOR);

            /*
            load all bundles that were installed last time the server was shut down
            */
            for (Iterator i= bundles.iterator(); i.hasNext();) {
                STBundle bundle= (STBundle)i.next();
                String name= bundle.getLocation();
                Exception x= null;
                STBundle stBundle= null;
                long bundleId= -1;
                try {
                    stBundle= installBundle(name);
                    bundleId= stBundle.getBundleId();
                }
                catch (Exception t) {
                    StringWriter stringWriter= new StringWriter();
                    t.printStackTrace(new PrintWriter(stringWriter));
                    m_log.log(LogService.LOG_ERROR,
                        "Could not load bundle '"+name+"'\n" + stringWriter.toString());
                }

                bundle.setBundleId(bundleId);
            }

            /*
            start all bundles that were active last time the server was shut down.
            */
            for (Iterator i= bundles.iterator(); i.hasNext();) {
                Bundle bundle= (Bundle)i.next();
                String name= bundle.getLocation();
                long bundleId= bundle.getBundleId();

                // skip bundles that failed to load
                if (bundleId < 0)
                  continue;

                STBundle stBundle= getBundle(bundleId);
                try {
                    if (stBundle != null  &&  bundle.getState() == Bundle.ACTIVE)
                        startBundle(bundleId);
                }
                catch (Exception t) {
                    m_log.log(LogService.LOG_ERROR, "Could not start bundle "+bundleId+":"+name, t);
                }
            }
            setState(STARTED);
        }
        finally {
            if (getState() != STARTED)
                setState(BEGIN);
        }
    }

    synchronized public void shutDown() {
        setState(STOPPING);
        try {
            for (Iterator i= m_bundles.iterator(); i.hasNext();) {
                STBundle bundle= (STBundle)i.next();
                long id= bundle.getBundleId();
                try {
                    if (bundle.getState() == Bundle.ACTIVE)
                        stopBundle(id);
                }
                catch (Exception t) {
                    StringWriter stringWriter= new StringWriter();
                    t.printStackTrace(new PrintWriter(stringWriter));
                    m_log.log(LogService.LOG_ERROR,
                        "Error stopping bundle '"+bundle.getLocation()+"'\n" + stringWriter.toString());
                }
            }
        }
        finally {
            setState(STOPPED);
        }
    }

    /**
     * returns the state of the Bundle Manager
     */
    protected int getState() {
        return m_state;
    }
    protected void setState(int state) {
        m_state= state;
    }

    synchronized public STBundle getBundle(long id) {
        STBundle bundle= (STBundle)m_bundleIDs.get(new Long(id));
        return bundle;
    }
    synchronized public STBundle getBundle(String location) {
        STBundle bundle= (STBundle)m_bundleNames.get(location);
        return bundle;
    }
    synchronized public STBundle[] getBundles() {
        Object[] a= m_bundles.toArray();
        STBundle[] b= new STBundle[a.length];
        System.arraycopy(a, 0, b, 0, a.length);
        return b;
    }
    synchronized long getOpenId() {
        Bundle[] bundles= getBundles();
        long id= 0;
        for (int i= bundles.length; 0 < i--;) {
            long j= bundles[i].getBundleId();
            if (id < j)
                id= j;
        }
        return ++id;
    }
    synchronized void putBundle(Bundle bundle) {
        if (m_bundles.contains(bundle))
            return;
        m_bundles.add(bundle);
        m_bundleIDs.put(new Long(bundle.getBundleId()), bundle);
        m_bundleNames.put(bundle.getLocation(), bundle);
    }
    synchronized void removeBundle(Bundle bundle) {
        m_bundles.remove(bundle);
        m_bundleNames.remove(bundle.getLocation());
        m_bundleIDs.remove(new Long(bundle.getBundleId()));
    }


    /**
     * Start this bundle.
     *
     * @preconditons  getState() in {INSTALLED,RESOLVED}.
     * @postconditons   if no exceptions thrown then getState() in {ACTIVE}.
     *
     * @throws
     *      BundleException - If the bundle couldn't be started.
     *      This could be because a code dependency could not be resolved or
     *      the specified BundleActivator could not be loaded or threw an
     *      exception.
     *
     *      java.lang.SecurityException - If the caller does not have the
     *      AdminPermission and the Java runtime environment supports
     *      permissions.
     */
    synchronized protected void startBundle(long bundleId)
    throws BundleException
    {
        STBundle stBundle= getBundle(bundleId);
        synchronized (stBundle) {
            int state= stBundle.getState();
            /*
            If the bundle is STARTING or STOPPING then this method will wait for
            the bundle to change state before continuing.
            */
            waitWhileStartingOrStopping(stBundle);

            /*
            If the bundle is ACTIVE then this method returns immediately.
            */
            if (state != Bundle.ACTIVE) {
                try {
                    /*
                    If the bundle is not RESOLVED, an attempt is made to resolve the bundle.
                    If the bundle cannot be resolved, a BundleException is thrown.
                    */
                    resolveBundleStartDependencies(stBundle);
                    stBundle.setState(Bundle.RESOLVED);

                    /*
                    The state of the bundle is set to STARTING.
                    */
                    stBundle.setState(Bundle.STARTING);
                    try {
                        /*
                        The start method of the bundle's BundleActivator, if one is specified,
                        is called. If the BundleActivator is invalid or throws an exception, the
                        state of the bundle is set back to RESOLVED, the bundle's listeners, if
                        any, are removed, service's registered by the bundle, if any, are
                        unregistered, and service's used by the bundle, if any, are released.
                        A BundleException is then thrown.
                        */
                        BundleActivator activator= stBundle.getBundleActivator();
                        BundleContext bundleContext= stBundle.getBundleContext();
                        if (activator != null) {
                            try {
                                activator.start(bundleContext);
                            }
                            catch (Throwable x) {
                                throw new STException("Failed to start bundle '"+stBundle.getLocation()+"'", x);
                            }
                        }

                        /*
                        It is recorded that this bundle has been started, so that when the
                        framework is restarted, this bundle will be automatically started.
                        */
                        try {
                            m_storage.updateBundleState(stBundle);
                        }
                        catch (Exception x) {
                            m_log.log(LogService.LOG_WARNING, "Failed to record that bundle '"+stBundle.getLocation()+"' was started.", x);
                        }
                    }

                    /*
                    if an error occurs while trying to start the bundle then
                    restore the state
                    */
                    catch (BundleException x) {
                        stBundle.setState(Bundle.RESOLVED);
                        throw new STException(x); // rethrow the exception
                    }

                    /*
                    The state of the bundle is set to ACTIVE.
                    */
                    stBundle.setState(Bundle.ACTIVE);

                    /*
                    remember that bundle was started
                    */
                    try {
                        m_storage.updateBundleState(stBundle);
                    }
                    catch (Throwable t) {
                        m_log.log(
                            LogService.LOG_WARNING,
                            "Failed to persist bundle info for Bundle "+stBundle.getBundleId()+":"+stBundle.getLocation(),
                            t);
                    }

                    /*
                    A BundleEvent of type BundleEvent.STARTED is broadcast.
                    */
                    fireBundleEvent(new BundleEvent(BundleEvent.STARTED, stBundle));
                }

                /*
                If an error occurs then release any packages or services that we
                may have locked.
                */
                catch (BundleException x) {
                    try { unregisterAllServices(stBundle); } catch (Throwable t) { }
                    try { releaseAllServices(stBundle); } catch (Throwable t) { }
                    try { releaseAllPackages(stBundle); } catch (Throwable t) { }
                    try { removeAllListeners(stBundle); } catch (Throwable t) { }
                    throw new STException(x); // rethrow the execption
                }
            }
        }
    }

    private void resolveBundleStartDependencies(STBundle stBundle)
    throws BundleException
    {
    }
    /**
     * Release all services that are being used by the denoted bundle.
     */
    private void releaseAllServices(STBundle stBundle)
    throws BundleException
    {
        for (Iterator i= m_registered.iterator(); i.hasNext();) {
            STServiceRegistration reg= (STServiceRegistration)i.next();
            STServiceReference ref= (STServiceReference)reg.getReference();
            Bundle bundle= ref.getBundle();
            if (bundle.equals(stBundle))
                i.remove();
        }
    }
    /**
     * Release all packages that are being used by the denoted bundle.
     */
    private void releaseAllPackages(STBundle stBundle)
    throws BundleException
    {
        /*
        There is really nothing to do here since removing the bundle from the
        manager's list of bundles also makes the bundle's packages unavailable
        from STClassLoader.
        */
    }
    /**
     * Remove all the listeners that were added by the denoted Bundle to other
     * system components.
     */
    private synchronized void removeAllListeners(STBundle stBundle)
    throws BundleException
    {
        EventListenerList.Entry[] entries= m_listeners.getListenerList();
        for (int i = 0; i < entries.length; i++) {
            ListenerEntry e= (ListenerEntry)entries[i];
            if (e.getBundle().equals(stBundle))
                m_listeners.remove(e);
        }
    }
    /**
     * Resolve the bundle's classpath and native code dependencies, if any.
     */
    private void resolveBundleInstallDependencies(STBundle stBundle)
    throws BundleException
    {
        /*
        The Framework makes available to the bundle its classpath JAR files and
        native code libraries that the bundle has specified in its
        Bundle-ClassPath and Bundle-NativeCode Manifest headers, respectively.
        The specified JAR files and native code libraries must be contained in
        the bundle's own JAR file.
        */
        // UNFINISHED
        // Assume classpath and code dependencies are available on the system path & classpath
    }

    final public STBundle installBundle(String location)
    throws BundleException
    {
        return installBundle(location, location, null, false/*don't force a re-install*/, true/*broadcast event*/);
    }
    final public STBundle installBundle(String location, InputStream in)
    throws BundleException
    {
        return installBundle(location, location, in, false/*don't force a re-install*/, true/*broadcast event*/);
    }
    synchronized protected STBundle installBundle(String location, String source, InputStream in, boolean forceReload, boolean broadcastEvent)
    throws BundleException
    {
        /*
        NOTES:
        Installing a bundle within the Framework must be a persistent operation:
        the bundle must remain installed across Java Virtual Machine invocations,
        until it is explicitly uninstalled.

        The installation process must be atomic: either this method completely
        installs the bundle or, if the installation fails, the Framework must be
        left in the state before this install method is called.

        Once a bundle is installed, an instance of Bundle is created; the rest
        of the lifecycle operations are performed against that Bundle object.
        */

        /*
        Check that the bundle is not already installed.
        If a bundle with the same location string has already been installed
        into the Framework, return that bundle.  The location string is used to
        uniquely identify an installed bundle.
        */
        STBundle[] bundles= getBundles();
        for (int i= 0; i < bundles.length; i++) {
            if (bundles[i].getLocation().equals(location))
                return bundles[i];
        }

        /*
        Check to see if the bundle has been persisted to storage but has not
        yet been loaded into memory.
        */
        STBundleContents contents= null;
        boolean bundleExists= false;
        if (!forceReload)
            if ((contents= m_storage.getBundleContents(location)) != null)
                bundleExists= true;

        /*
        Obtain the bundle's contents.
        If an InputStream is specified, the Framework will try to fetch the
        bundle from the supplied stream.  Otherwise, the Framework will try to
        download the bundle using its location string, which typically is the
        string representation of a URL identifying the bundle's codebase.

        Persist the bundle's contents the bundle's associated data directory.
        */
        if (!bundleExists)
            contents= m_storage.installBundle(location, source, in);

        STClassLoader classLoader= new STClassLoader(this, location, contents);
        STBundle stBundle= classLoader.createBundle();
        long id= stBundle.getBundleId();

        /*
        Set the bundle's state to INSTALLED.
        */
        stBundle.setState(Bundle.INSTALLED);

        /*
        put bundle in internal lists.
        */
        putBundle(stBundle);
        try {
            /*
            Resolve the bundle's classpath and native code dependencies, if any.
            The Framework makes available to the bundle its classpath JAR files
            and native code libraries that the bundle has specified in its
            Bundle-ClassPath and Bundle-NativeCode Manifest headers.
            The specified JAR files and native code libraries must be contained in
            the bundle's own JAR file.
            */
            resolveBundleInstallDependencies(stBundle);

            stBundle.setState(Bundle.RESOLVED);

            try {
                // remember that bundle was installed
                m_storage.updateBundleState(stBundle);
            }
            catch (Throwable t) {
                m_log.log(
                    LogService.LOG_WARNING,
                    "Failed to persist bundle info for Bundle "+stBundle.getBundleId()+":"+stBundle.getLocation(),
                    t);
            }

            /*
            Broadcast a BundleEvent of type BundleEvent.INSTALLED.
            */
            fireBundleEvent(new BundleEvent(BundleEvent.INSTALLED, stBundle));
        }
        catch (BundleException x) {
            uninstallBundle(id, false/*don;t broadcast event*/);
        }

        return stBundle;
    }
    public void uninstallBundle(long bundleId) {
        uninstallBundle(bundleId, true);
    }
    /**
     * Uninstalls a bundle.
     * Once this method returns, the Framework state is the same as if the
     * bundle had never been installed.
     */
    synchronized protected void uninstallBundle(long bundleId, boolean broadcast)
    {
        STBundle stBundle= getBundle(bundleId);
        if (stBundle == null)
            throw new IllegalStateException("Unknown bundle Id");

        int state= stBundle.getState();
        if (state == Bundle.ACTIVE) {
            try {
                stopBundle(bundleId);
            }
            catch (BundleException x) {
                String msg= "Failed to stop bundle "+bundleId+":"+stBundle.getLocation()+" before uninstalling it.";
                m_log.log(LogService.LOG_WARNING, msg, x);
            }
        }

        if (broadcast)
            fireBundleEvent(new BundleEvent(BundleEvent.UNINSTALLED, stBundle));

        stBundle.setState(Bundle.UNINSTALLED);

        try {
            m_storage.uninstallBundle(stBundle.getLocation());
        }
        catch (Exception x) {
            m_log.log(LogService.LOG_WARNING, "Error uninstalling bundle from storage.", x);
        }
        removeBundle(stBundle);
    }

    synchronized protected void stopBundle(long id)
    throws BundleException
    {
        STBundle stBundle= getBundle(id);
        if (stBundle == null)
            throw new IllegalStateException("Unknown bundle Id");
        synchronized (stBundle) {
            /*
            If the bundle is UNINSTALLED then an IllegalStateException is thrown.
            */
            int state= stBundle.getState();
            if (state == Bundle.UNINSTALLED)
                throw new IllegalStateException("Bundle '"+stBundle.getBundleId()+"' is UNINSTALLED");

            /*
            If the bundle is STARTING or STOPPING then this method will wait for
            the bundle to change state before continuing.
            */
            waitWhileStartingOrStopping(stBundle);

            /*
            If the bundle is not ACTIVE then this method returns immediately.
            */
            if (state == Bundle.ACTIVE) {
                /*
                The state of the bundle is set to STOPPING.
                */
                stBundle.setState(Bundle.STOPPING);

                /*
                The stop method of the bundle's BundleActivator, if one is specified,
                is called. If the BundleActivator is invalid or throws an exception, the
                state of the bundle is set back to RESOLVED, the bundle's listeners, if
                any, are removed, service's registered by the bundle, if any, are
                unregistered, and service's used by the bundle, if any, are released.
                A BundleException is then thrown.
                */
                BundleActivator activator= stBundle.getBundleActivator();
                BundleContext bundleContext= stBundle.getBundleContext();
                if (activator != null) {
                    try {
                        activator.stop(bundleContext);
                    }
                    catch (Throwable x) {
                        m_log.log(LogService.LOG_WARNING, "Error while stopping bundle '"+stBundle.getLocation()+"'.", x);
                    }
                }

                /*
                Unregister any event listeners and remaining services
                registered by this bundle; release any services used by
                this bundle.
                */
                try {
                    unregisterAllServices(stBundle);
                }
                catch (BundleException x) {
                    m_log.log(LogService.LOG_WARNING, "Error while unregistering services for bundle '"+stBundle.getLocation()+"'.", x);
                }
                try {
                    removeAllListeners(stBundle);
                }
                catch (BundleException x) {
                    m_log.log(LogService.LOG_WARNING, "Error while removing listeners from bundle '"+stBundle.getLocation()+"'.", x);
                }
                try {
                    releaseAllServices(stBundle);
                }
                catch (BundleException x) {
                    m_log.log(LogService.LOG_WARNING, "Error while releasing services for bundle '"+stBundle.getLocation()+"'.", x);
                }
                try {
                    releaseAllPackages(stBundle);
                }
                catch (BundleException x) {
                    m_log.log(LogService.LOG_WARNING, "Error while releasing packages for bundle '"+stBundle.getLocation()+"'.", x);
                }

                stBundle.setState(Bundle.RESOLVED);

                /*
                persist the bundle's state so that when the
                framework is restarted, this bundle will not be
                automatically started.

                if the bundle manager is stopping then we do not record/persist
                the state change since we want bundles to start up again the next
                time the framework is run
                */
                if (getState() != STOPPING) {
                    try {
                        m_storage.updateBundleState(stBundle);
                    }
                    catch (Exception x) {
                        m_log.log(LogService.LOG_WARNING, "Failed to record that bundle '"+stBundle.getLocation()+"' was stopped.", x);
                    }
                }

                /*
                A BundleEvent of type BundleEvent.STOPPED is broadcast.
                */
                fireBundleEvent(new BundleEvent(BundleEvent.STOPPED, stBundle));
            }
        }
    }

    /**
     * If the bundle is STARTING or STOPPING then this method will wait for
     * the bundle to change state before continuing. If this does not occur
     * in a reasonable time, a BundleException is thrown to indicate the
     * bundle was unable to be started.
     */
    private void waitWhileStartingOrStopping(STBundle stBundle)
    throws BundleException
    {
        synchronized (stBundle) {
            int state= stBundle.getState();
            if (state == Bundle.STARTING  ||  state == Bundle.STOPPING) {
                long timeout= m_configuration.getBundleTimeoutMinutes() * 60 * 1000;
                long start= System.currentTimeMillis();
                for (;state == Bundle.STARTING  ||  state == Bundle.STOPPING;) {
                    try {
                        stBundle.wait(10 * 1000);
                    }
                    catch (InterruptedException x) {
                    }
                    state= stBundle.getState();
                    if (start + timeout <= System.currentTimeMillis())
                        break;
                }
            }
            if (state == Bundle.STARTING  ||  state == Bundle.STOPPING)
                throw new STException("Time out while waiting for bundle '"+stBundle.getLocation()+"'  to start or stop.");
        }
    }

    public void updateBundle(long bundleId)
    throws BundleException
    {
        updateBundle(bundleId, null);
    }
    public void updateBundle(long bundleId, InputStream in)
    throws BundleException
    {
        STBundle stBundle= getBundle(bundleId);
        if (stBundle == null)
            throw new IllegalStateException("Unknown bundle Id:"+bundleId);
        String location= stBundle.getLocation();

        /*
        Check the bundle's state:
        If the bundle is ACTIVE, stop it as described in the Bundle.stop method;
        if this throws an exception, rethrow it and return.
        If the bundle is UNINSTALLED, throw an IllegalStateException.
        */
        int beginState= stBundle.getState();
        if (beginState == Bundle.UNINSTALLED)
            throw new IllegalStateException("Bundle "+bundleId+" has been uninstalled");

        /*
        Back up the bundle's JAR file.
        */
        try {
            m_storage.backupBundle(location);
        }
        catch (IOException x) {
            throw new BundleException("Error updating bundle "+location+", could not backup bundle", x);
        }

        /*
        If the bundle is ACTIVE, stop it as described in the Bundle.stop method;
        */
        if (beginState == Bundle.ACTIVE) {
            try {
                stopBundle(stBundle.getBundleId());
            }
            catch (BundleException x) {
                throw new STException("Error during update, could not stop the bundle", x);
            }
        }

        /*
        Install the new version of the bundle.
        If an InputStream is supplied, get the bundle from the stream.
        Otherwise, if the bundle has a Bundle-UpdateLocation Manifest header,
        use its value to retrieve the new bundle. Otherwise, call the bundle's
        getLocation method to determine its original location and retrieve the
        new version from there.
        */
        boolean success= false;
        try {
            String source= null;
            Dictionary headers= stBundle.getHeaders();
            String updateLocation= (String)headers.get("Bundle-UpdateLocation");
            if (updateLocation != null  &&  0 < updateLocation.trim().length())
                source= updateLocation;
            if (source == null)
                source= location;
            STBundle updatedBundle=
                installBundle(
                    location,
                    source,
                    in,
                    true/*force reinstall*/,
                    false/*don't brodacast install event*/);

            /*
            Resume the target bundle if it is initially active.
            This step performs the equivalent of the Bundle.start method.
            */
            if (beginState == Bundle.ACTIVE)
                startBundle(bundleId);

            /*
            If the update process has succeeded: Broadcast a BundleEvent of type
            BundleEvent.UPDATED by calling the bundleChanged method of each
            registered BundleListener.
            */
            BundleEvent evt= new BundleEvent(BundleEvent.UPDATED, updatedBundle);
        }
        finally {
            if (!success) {
                /*
                If the update process has failed: Revert the bundle to its old version.
                */
                try { m_storage.restoreBundle(location); } catch (Throwable t) { }

                /*
                Resume the given bundle if it is initially active.
                This step performs the equivalent of the Bundle.start method.
                */
                if (beginState == Bundle.ACTIVE) {
                    try { startBundle(bundleId); } catch (Throwable t) { }
                }
            }
        }
    }

    synchronized public ServiceRegistration registerService(long bundleId, String[] clazzes, Object service, Dictionary properties)
    {
        if (clazzes == null)
            clazzes= new String[0];

        if (properties == null)
            properties= new Hashtable(0);

        STBundle bundle= getBundle(bundleId);
        if (bundle == null)
            throw new IllegalArgumentException("Unknow bundle Id");

        if (service == null)
            throw new IllegalArgumentException("The service parameter is null");

        if (!(service instanceof ServiceFactory)) {
            Class serviceClass= service.getClass();
            ClassLoader classLoader= bundle.getClassLoader();
            for (int i= 0; i < clazzes.length; i++) {
                try {
                    Class clazz= classLoader.loadClass(clazzes[i]);
                    if (!clazz.isAssignableFrom(serviceClass))
                        throw new IllegalArgumentException("The service parameter is not an instanceof all the named classes in the clazzes parameter");
                }
                catch (Throwable t) {
                    throw new IllegalArgumentException("Could not load class "+clazzes[i]);
                }
            }
        }

        /*
        A bundle may only register a service while the Framework is activating
        the bundle or at any time during its active phase.
        */
        int state= bundle.getState();
        if (state != Bundle.ACTIVE && state != Bundle.STARTING)
            throw new IllegalStateException("A bundle may only register a service while the Framework is activating the bundle or at any time during its active phase.");

        STServiceReference reference= new STServiceReference();
        STServiceRegistration registration= new STServiceRegistration(reference);
        reference.init(this, registration, bundle, clazzes, service, properties);

        m_registered.add(registration);

        return registration;
    }
    /**
     * @return  true if the given reference is associated with a registered service
     */
    synchronized protected boolean isRegisteredService(STServiceReference reference) {
        for (Iterator i= m_registered.iterator(); i.hasNext();) {
            STServiceRegistration registration= (STServiceRegistration)i.next();
            if (registration.getReference().equals(reference))
                return true;
        }
        return false;
    }
    /**
     * @param   bundle  the bundle that registered the service
     */
    synchronized protected void unregisterService(STServiceRegistration registration) {
        STServiceReference reference= (STServiceReference)registration.getReference();
        if (!m_registered.contains(registration))
            throw new IllegalStateException("The service is not currently registered");

        ServiceEvent evt= new ServiceEvent(ServiceEvent.UNREGISTERING, reference);
        fireServiceEvent(evt);

        m_registered.remove(registration);
        reference.unregisterService();
    }
    /**
     * Unregister all services that were registered by the denoted bundle.
     */
    synchronized private void unregisterAllServices(STBundle stBundle)
    throws BundleException
    {
        HashSet registered= new HashSet(m_registered);
        for (Iterator i= registered.iterator(); i.hasNext();) {
            STServiceRegistration reg= (STServiceRegistration)i.next();
            STServiceReference sr= (STServiceReference)reg.getReference();
            if (sr.getBundle().equals(stBundle)) {
                try {
                    unregisterService(reg);
                }
                catch (Throwable t) {
                    m_log.log(LogService.LOG_WARNING, "Failed to unregister services "+sr.getClasses(), t);
                }
            }
        }
    }
    /**
     * @return a list of registered services
     */
    synchronized public ServiceReference[] getRegisteredServices() {
        ServiceReference[] registered= new ServiceReference[m_registered.size()];
        int j= 0;
        for (Iterator i= m_registered.iterator(); i.hasNext();) {
            STServiceRegistration reg= (STServiceRegistration)i.next();
            registered[j++]= reg.getReference();
        }
        return registered;
    }
    public ServiceReference[] getServicesInUse(long bundleId) {
        ArrayList refs= new ArrayList();
        for (Iterator i= m_registered.iterator(); i.hasNext();) {
            STServiceRegistration reg= (STServiceRegistration)i.next();
            STServiceReference ref= (STServiceReference)reg.getReference();
            if (0 < ref.getUseCount(bundleId))
                refs.add(ref);
        }
        ServiceReference[] used= new ServiceReference[refs.size()];
        refs.toArray(used);
        return used;
    }
    public ServiceReference[] getServiceReferences(String clazz, String filter)
    throws InvalidSyntaxException
    {
        ArrayList refs= new ArrayList();
        for (Iterator i= m_registered.iterator(); i.hasNext();) {
            STServiceRegistration reg= (STServiceRegistration)i.next();
            STServiceReference ref= (STServiceReference)reg.getReference();
            String[] clazzes= ref.getClasses();
            for (int j= 0; j < clazzes.length; j++) {
                if (clazzes[j].equals(clazz))
                    if (filter != null) {
                        Dictionary properties= ref.getProperties();
                        if (LDAPUtils.match(properties, filter)) {
                            refs.add(ref);
                            break;
                        }
                    }
                    else {
                        refs.add(ref);
                        break;
                    }
            }
        }
        ServiceReference[] used= new ServiceReference[refs.size()];
        refs.toArray(used);
        return used;
    }
    public ServiceReference getServiceReference(String clazz)
    {
        ServiceReference[] refs= null;
        try {
            refs= getServiceReferences(clazz, null);
        }
        catch (InvalidSyntaxException x) { /* not gonna happen */ }
        if (refs == null  ||  refs.length <= 0)
            return null;
        return refs[0];
    }
    public Object getService(long bundleId, STServiceReference reference) {
        Object service= null;
        Long id= new Long(bundleId);
        /*
        If the service has been unregistered, null is returned.
        */
        if (!isRegisteredService(reference))
            return null;

        try {
            service= reference.getService(bundleId);
        }
        /*
        If the ServiceFactory throws an exception, null is returned and a
        FrameworkEvent of type FrameworkEvent.ERROR is broadcast.
        */
        catch (Throwable t) {
            FrameworkEvent e= new FrameworkEvent(FrameworkEvent.ERROR, reference.getBundle(), t);
            fireFrameworkEvent(e);
            service= null;
        }

        return service;
    }
    public boolean ungetService(long bundleId, STServiceReference reference) {
        if (!isRegisteredService(reference))
            return false;

        return reference.ungetService(bundleId);
    }

    public void addServiceListener(long bundleId, ServiceListener listener, String filter)
    throws InvalidSyntaxException
    {
        LDAPUtils.parse(filter);
        STBundle bundle= getBundle(bundleId);
        if (bundle == null)
            return;
        ListenerEntry entry= new ListenerEntry(ServiceListener.class, listener, bundle, filter);
        m_listeners.add(entry);
    }
    public void addServiceListener(long bundleId, ServiceListener listener) {
        try {
            addServiceListener(bundleId, listener, null);
        }
        catch (InvalidSyntaxException x) { /* not gonna happen */ }
    }
    public void removeServiceListener(ServiceListener l) {
        m_listeners.remove(ServiceListener.class, l);
    }
    public void fireServiceEvent(ServiceEvent e)
    {
        EventListenerList.Entry[] entries= m_listeners.getEntries(ServiceListener.class);
        for (int i = 0; i < entries.length; i++) {
            ListenerEntry entry= (ListenerEntry)entries[i];
            String filter= entry.getFilter();
            if (filter != null) {
                STServiceReference ref= (STServiceReference)e.getServiceReference();
                Dictionary properties= ref.getProperties();
                try {
                    if (!LDAPUtils.match(properties, filter))
                        continue;
                }
                catch (InvalidSyntaxException x) {
                }
            }
            ((ServiceListener)entry.getListener()).serviceChanged(e);
        }
    }

    public void addFrameworkListener(long bundleId, FrameworkListener listener) {
        STBundle bundle= getBundle(bundleId);
        if (bundle == null)
            return;
        ListenerEntry entry= new ListenerEntry(FrameworkListener.class, listener, bundle, null);
        m_listeners.add(entry);
    }
    public void removeFrameworkListener(FrameworkListener listener) {
        m_listeners.remove(FrameworkListener.class, listener);
    }
    public void fireFrameworkEvent(FrameworkEvent event)
    {
        Object[] listeners = m_listeners.getListenerList(FrameworkListener.class);
        for (int i = 0; i < listeners.length; i++) {
                ((FrameworkListener)listeners[i]).frameworkEvent(event);
        }
    }
    public void addBundleListener(long bundleId, BundleListener listener) {
        STBundle bundle= getBundle(bundleId);
        if (bundle == null)
            return;
        ListenerEntry entry= new ListenerEntry(BundleListener.class, listener, bundle, null);
        m_listeners.add(entry);
    }
    public void removeBundleListener(BundleListener listener) {
        m_listeners.remove(BundleListener.class, listener);
    }
    public void fireBundleEvent(BundleEvent event)
    {
        Object[] listeners = m_listeners.getListenerList(BundleListener.class);
        for (int i = 0; i < listeners.length; i++) {
                ((BundleListener)listeners[i]).bundleChanged(event);
        }
    }
}
