/*
   STBundle.java

   Copyright (C) 2000 by ted stockwell
   All Rights Reserved.
 */
package org.servicetango.core;

import org.servicetango.util.*;
import org.osgi.framework.*;
import java.util.AbstractList;

import java.io.*;
import java.util.*;
import java.util.jar.*;

/**
 * ServiceTango's implementation of the Bundle interface.
 * Beside serving as a bundle's interface to the Framework, an STBundle also
 * serves as a repository (producer and consumer) of data specific to the
 * associated bundle.
 *
 * @author   ted stockwell [emorning@mediaone.net]
 */
public class STBundle
extends ProfileMap
implements Bundle
{
    public static final String KEY_STATE= "state";
    public static final String KEY_LOCATION= "location";
    public static final String KEY_BUNDLE_ID= "bundle_id";

    public static final String PROPERTY_STATE= STBundle.class.getName()+".state";

    STBundleContents m_contents= null;
    Manifest m_manifest= null;
    Set m_registeredServices= new HashSet();
    Set m_usedServices= new HashSet();
    Collection m_exportPackages= null;
    BundleActivator m_activator= null;
    BundleContext m_context= null;
    BundleManager m_manager= null;
    STClassLoader m_classLoader= null;

    /**
     * STBundle are created by calling the newInstance method of a custom class
     * loader.  Therefore an emprty constructor is required for this class and
     * any descendants of this class.
     */
    public STBundle() {
    }
    /**
     * Initializes this object.
     * This method must be called before calling any other method in this class.
     */
    synchronized public void init(long id, STClassLoader classLoader)
    throws BundleException
    {
        m_classLoader= classLoader;
        setBundleId(id);
        setLocation(classLoader.getLocation());
        m_contents= classLoader.getBundleContents();
        m_manifest= m_contents.getManifest();
        m_manager= classLoader.getBundleManager();
    }

    public STBundleContents getBundleContents() {
        return m_contents;
    }
    public BundleManager getBundleManager() {
        return m_manager;
    }
    public STClassLoader getClassLoader() {
        return m_classLoader;
    }

    public BundleActivator getBundleActivator()
    throws BundleException
    {
        if (m_activator == null) {
            synchronized (this) {
                if (m_activator == null) {
                    Dictionary headers= getHeaders();
                    String name= (String)headers.get("Bundle-Activator");
                    if (name == null)
                        return null;
                    try {
                        Class clazz= m_classLoader.loadClass(name);
                        m_activator= (BundleActivator)clazz.newInstance();
                        return m_activator;
                    }
                    catch (Throwable t) {
                        throw new STException("Error loading activator for bundle '"+getLocation()+"'", t);
                    }
               }
           }
        }
        return m_activator;
    }

    protected BundleContext getBundleContext() {
        if (m_context == null) {
            synchronized (this) {
                if (m_context == null)
                    m_context= new STBundleContext(this);
           }
        }
        return m_context;
    }

    ///////////////////////////////////////
    //
    // methods from Bundle interface
    //
    public synchronized int getState() {
        return getInt(KEY_STATE, Bundle.UNINSTALLED);
    }
    synchronized public void start()
    throws BundleException
    {
        /*
        if the bundle is already started then we're done
        */
        int state= getState();
        if (state == Bundle.ACTIVE)
            return;

        /*
        If the bundle is UNINSTALLED then an IllegalStateException is thrown.
        */
        if (state == Bundle.UNINSTALLED)
            throw new IllegalStateException("Bundle '"+getLocation()+"' is UNINSTALLED");

        m_manager.startBundle(getBundleId());

        setState(Bundle.ACTIVE);
    }
    synchronized public void stop()
    throws BundleException
    {
        int state= getState();
        if (state != Bundle.ACTIVE)
            return;
        m_manager.stopBundle(getBundleId());
        setState(Bundle.RESOLVED);
    }
    synchronized public void update()
    throws BundleException
    {
        m_manager.updateBundle(getBundleId());
    }
    synchronized public void update(InputStream in)
    throws BundleException
    {
        m_manager.updateBundle(getBundleId(), in);
    }
    synchronized public void uninstall()
    throws BundleException
    {
        m_manager.uninstallBundle(getBundleId());
    }
    public Dictionary getHeaders() {
        Map m= m_manifest.getMainAttributes();
        CaselessProfile profile= new CaselessProfile(m.size());
        for (Iterator i= m.entrySet().iterator(); i.hasNext();) {
            Map.Entry entry= (Map.Entry)i.next();
            String key= ((Attributes.Name)entry.getKey()).toString();
            String value= entry.getValue().toString();
            profile.put(key, value);
        }
        Dictionary d= CollectionUtils.toDictionary(profile);
        return d;
    }
    public long getBundleId() {
        return this.getLong(KEY_BUNDLE_ID, -1);
    }
    public String getLocation() {
        return getString(KEY_LOCATION);
    }
    synchronized public ServiceReference[] getRegisteredServices() {
        return m_manager.getRegisteredServices();
    }
    synchronized public ServiceReference[] getServicesInUse() {
        return m_manager.getServicesInUse(getBundleId());
    }
    public boolean hasPermission(Object p0) {
        // UNFINISHED
        return false;
    }

    protected void setBundleId(long id) {
        setLong(KEY_BUNDLE_ID, id);
    }
    protected void setLocation(String location) {
        put(KEY_LOCATION, location);
    }
    protected void setState(int state) {
        setInt(KEY_STATE, state);
    }

    public boolean equals(Object o) {
        if (o instanceof Bundle)
            return getLocation().equals(((Bundle)o).getLocation());
        if (getLocation().equals(o.toString()))
            return true;
        return super.equals(o);
    }
    public int hashCode() {
        return getLocation().hashCode();
    }
}
