/*
   STClassLoader.java

   Copyright (C) 2000 by ted stockwell
   All Rights Reserved.
 */
package org.servicetango.core;

import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.jar.*;
import java.util.zip.*;
import java.text.SimpleDateFormat;

import org.osgi.framework.*;
import org.osgi.service.log.*;


/**
 * A custom classloader for a Bundle that will load classes and resources from...
 *      - the bundle's jar file
 *      - exported classes from other jar files that are in turn imported by
 *          the given bundle.
 *      - the system classpath
 * ...in this order.
 *
 * <P>There is no way for a classloader to 'undefine' a class once it
 * has been loaded.  However, the servlet engine can discard a
 * classloader and the classes it contains, causing the
 *
 * @author ted stockwell [emorning@mediaone.net]
 * @version $Revision$
 **/
public class STClassLoader
extends ClassLoader
{
    BundleManager m_manager;
    STBundleContents m_jar;
    String m_location;

    /**
     * Cache of the loaded classes. This contains ClassCacheEntry keyed
     * by class names.
     */
    private Map m_cache= new HashMap();

    /**
     * Creates a class loader for the given bundle
     */
    public STClassLoader(BundleManager manager, String location, STBundleContents jar) {
        m_manager= manager;
        m_jar= jar;
        m_location= location;
    }

    public String getLocation() {
        return m_location;
    }
    public BundleManager getBundleManager() {
        return m_manager;
    }
    public STBundleContents getBundleContents() {
        return m_jar;
    }

    protected STBundle createBundle()
    throws BundleException
    {
        try {
            Class bundleClass= loadClass("org.servicetango.core.STBundle");
            STBundle stBundle= (STBundle)bundleClass.newInstance();
            stBundle.init(m_manager.getOpenId(), this);
            return stBundle;
        }
        catch (ClassNotFoundException x) {
            throw new STException("Internal Error", x);
        }
        catch (InstantiationException x) {
            throw new STException("Internal Error", x);
        }
        catch (IllegalAccessException x) {
            throw new STException("Internal Error", x);
        }
    }


   //------------------------------------ Implementation of Classloader
    /**
     * Implementation of ClassLoader.loadClass(String name, boolean resolve)
     *
     * @see ClassLoader.loadClass
     **/
    protected synchronized Class loadClass(String name, boolean resolve)
    throws ClassNotFoundException
    {
        if (!securityAllowsClass(name))
            return getClass().getClassLoader().loadClass(name);

        /*
        Check cache
        */
        Class clazz= (Class)m_cache.get(name);
        if (clazz != null)
            return clazz;

        /*
        Check the bundle's jar file
        */
	//Translate class name to file name
	String classFileName = name.replace('.', '/') + ".class";
        byte[] classData= m_jar.getEntryData(classFileName);
        if (classData != null)
            clazz= defineClass(name, classData, 0, classData.length);

        /*
        try loading the class from the system classpath
        */
        if (clazz == null) {
            try {
                clazz= getClass().getClassLoader().loadClass(name);
                resolve= false; // already resolved
            }
            catch (Throwable t) {
            }
        }

        /*
        check the exported packages from other bundles
        */
        if (clazz == null) {
            STBundle[] bundles= m_manager.getBundles();
            String packageName= name.substring(0, name.lastIndexOf('.'));
            for (int i= 0; i < bundles.length; i++) {
                STBundle bundle= bundles[i];
                STBundleContents contents= bundle.getBundleContents();
                Collection exportedPackages= contents.getExportedPackageNames();
                if (exportedPackages.contains(packageName)) {
                    classData= contents.getEntryData(classFileName);
                    if (classData != null)
                        clazz= defineClass(name, classData, 0, classData.length);
                }
            }
        }

        if (clazz != null) {
            //Resolve it if necessary
            if (resolve)
                resolveClass(clazz);
            m_cache.put(name, clazz);  //Cache the result;
            return clazz;
        }

        throw new ClassNotFoundException(name);
    }

    /** Checks whether a classloader is allowed to define a given class,
     * within the security manager restrictions. **/
    private boolean securityAllowsClass(String className) {
	try {
	    SecurityManager security = System.getSecurityManager();
	    if (security == null)
		return true;

	    int lastDot = className.lastIndexOf('.');
	    String pkgName =
		(lastDot > -1) ? className.substring(0, lastDot) : "";
	    security.checkPackageDefinition(pkgName);
	    // Throws if not allowed
	    return true;
	}
	catch (SecurityException e) {
	    return false;
	}
    }


    /**
     * Loads all the bytes of an InputStream.
     */
    private byte[] loadBytesFromStream(InputStream in, int length)
         throws IOException
     {
	byte[] buf = new byte[length];
	int nRead, count = 0;
	while ((length > 0) && ((nRead = in.read(buf,count,length)) != -1)) {
            count += nRead;
            length -= nRead;
	}
        return buf;
    }

    /**
     * Get an InputStream on a given resource.  Will return null if no
     * resource with this name is found.
     *
     * <P>The JavaClassLoader translate the resource's name to a file
     * or a zip entry. It looks for the resource in all its repository
     * entry.
     *
     * @see     java.lang.Class#getResourceAsStream(String)
     * @param	name	the name of the resource, to be used as is.
     * @return	an InputStream on the resource, or null if not found.
    public InputStream getResourceAsStream(String name) {
	//Try to load it from the system class
	InputStream s = getSystemResourceAsStream(name);
	if          (s == null) {
	    //Try to find it from every repository
	    Enumeration repEnum = repository.elements();
	    while       (repEnum.hasMoreElements()) {
		File file = (File) repEnum.nextElement();
		if   (file.isDirectory()) {
		    s = loadResourceFromDirectory(file, name);
		}
		else {
		    s = loadResourceFromZipfile(file, name);
		}
		if (s != null) {
		    break;	//Terminate search

		}
	    }
	}
	return s;
    }
     */

    /**
     * Loads resource from a directory.
     */
    private InputStream loadResourceFromDirectory(File dir, String name) {
	//Name of resources are always separated by /
	String fileName = name.replace('/', File.separatorChar);
	File resFile = new File(dir, fileName);
	if (resFile.exists()) {
	    try {
		return new FileInputStream(resFile);
	    } catch (FileNotFoundException shouldnothappen) {
		shouldnothappen.printStackTrace();
		return null;
	    }
	} else {
	    return null;
	}
    }

    /**
     * Loads resource from a zip file
     */
    private InputStream loadResourceFromZipfile(File file, String name) {
	try {
	    ZipFile zipfile = new ZipFile(file);
	    ZipEntry entry = zipfile.getEntry(name);
	    if (entry != null) {
		return zipfile.getInputStream(entry);
	    } else {
		return null;
	    }
	} catch(IOException ioe) {
	    return null;
	}
    }

    /**
     * Find a resource with a given name.  The return is a URL to the resource.
     * Doing a getContent() on the URL may return an Image, an AudioClip,
     * or an InputStream.<p>
     *
     * The JavaClassLoader looks for the resource only in the directory
     * repository for this resource.
     *
     * @param	name	the name of the resource, to be used as is.
     * @return	an URL on the resource, or null if not found.
     * @since   JDK1.1
    public java.net.URL getResource(String name) {
	URL u = getSystemResource(name);
	if  (u != null) {
	    return u;
	}
	//Load for it only in directories since no URL can point into
	//a zip file.

	Enumeration repEnum = repository.elements();
	while (repEnum.hasMoreElements()) {
	    File file = (File) repEnum.nextElement();
	    if (file.isDirectory()) {
		String fileName = name.replace('/', File.separatorChar);
		File resFile = new File(file, fileName);
		if (resFile.exists()) {
		    //Build a file:// URL form the file name
		    try {
			return new URL("file://" + resFile.getAbsolutePath());
		    } catch(java.net.MalformedURLException badurl) {
			badurl.printStackTrace();
			return null;
		    }
		}
	    }
	}
	//Not found
	return null;
    }
     */

    /**
     * Import  all packages denoted in the bundle's manifest.
     */
    public void importPackages()
    throws BundleException
    {
        /*
        FIXME

        Actually, there is nothing to do here since this class ignores what a bundle
        explicitly imports and basically imports all packages that are exported
        by other bundles.  I suppose I could make this class more strict with
        regard to what packages are imported but I'm not sure why I should put
        any effort into it.  It seems more friendly this way.
        */
    }

    /**
     * Export all packages denoted in the bundle's manifest.
     */
    public void exportPackages()
    throws BundleException
    {
        /*
        FIXME
        */
    }
}
