/*
   STStorage.java

   Copyright (C) 2000 by ted stockwell
   All Rights Reserved.
 */
package org.servicetango.core;

import org.servicetango.util.*;

import java.io.*;
import java.util.*;
import java.util.jar.*;
import java.net.*;

import org.osgi.framework.*;
import org.w3c.dom.*;
import org.xml.sax.*;

/**
 *  Manages a BundleManager's persistant storage.
 *
 *  @author  ted stockwell [emorning@mediaone.net]
 */
public class STStorage
{
    static final String BUNDLE_FOLDER_PREFIX= "bundle.";
    static final String FILE_SEPARATOR= System.getProperty("file.separator");
    static final String STORAGE_DIRECTORY_FILE_NAME= "storage.xml";

    public static final String KEY_BUNDLES= "bundle";


    BundleManager m_manager;
    String m_rootName;
    File m_root; // root directory

    // list of installed bundles
    ElementMap m_index= null;

    public STStorage(BundleManager manager)
    throws BundleException
    {
        m_manager= manager;
        m_rootName= m_manager.getConfiguration().getRootDirectory();
        m_root= new File(m_rootName);
        if (!m_root.exists()) {
            if (!m_root.mkdirs())
                throw new STException("Could not create storage directory '"+m_rootName+"'");
        }
        if (!m_root.isDirectory())
                throw new STException("Invalid storage directory name '"+m_rootName+"'");
    }

    public File getRootDirectory() {
        return m_root;
    }

    public String getIndexName() {
        File root= getRootDirectory();
        String directoryName= root.getAbsolutePath()+FILE_SEPARATOR+STORAGE_DIRECTORY_FILE_NAME;
        return directoryName;
    }

    synchronized protected ElementMap getIndex()
    throws IOException, SAXException
    {
        if (m_index != null)
            return m_index;

        InputStream in= null;
        try {
            String indexName= getIndexName();
            File fileIndex= new File(indexName);
            if (!fileIndex.exists()) {
                m_index= new ElementMap(new java.io.ByteArrayInputStream(("<storage></storage>").getBytes()));
            }
            else {
                in= new FileInputStream(indexName);
                in= new BufferedInputStream(in);
                m_index= new ElementMap(in);
            }
        }
        finally {
            if (in != null) {
                try { in.close(); } catch (Throwable t) { }
            }
        }
        return m_index;
    }

    synchronized protected void saveIndex(ElementMap index)
    throws IOException
    {
        if (index == null)
            return;
        String indexName= getIndexName();
        OutputStream out= new FileOutputStream(indexName);
        try {
            out= new BufferedOutputStream(out);
            index.save(out);
        }
        finally {
            if (out != null) {
                try { out.close(); } catch (Throwable t) { }
            }
        }
    }

    /**
     * Update bundle information, like the state of the bundle, of a previously
     * installed bundle.
     */
    public void updateBundleState(STBundle stBundle)
    throws IOException, SAXException
    {
        ElementMap index= getIndex();
        List bundles= getStorageElements(index);

        // remove the bundle
        String location= stBundle.getLocation();
        for (int i= bundles.size(); 0 < i--;) {
            Map m= (Map)bundles.get(i);
            String l= LangUtils.toString(m.get(STBundle.KEY_LOCATION));
            if (l != null  &&  l.equals(location))
                bundles.remove(i);
        }

        // add bundle
        bundles.add(stBundle);
        putBundleElements(index, bundles);
        saveIndex(index);
    }

    void putBundleElements(Map index, List bundles) {
        index.put(KEY_BUNDLES, bundles);
    }

    public List getStorageElements(Map index)
    throws SAXException, IOException
    {
        List bundles;
        Object value= index.get(KEY_BUNDLES);
        if (value == null) {
            bundles= new ArrayList(0);
        }
        else if (value instanceof List) {
            bundles= (List)value;
        }
        else {
            bundles= new ArrayList(1);
            bundles.add(value);
        }
        return bundles;
    }

    /**
     * Return a list of the bundles that have been persisted to storage.
     * @returns  A List of Bundle objects.
     */
    protected Set getInstalledBundles()
    throws SAXException, IOException
    {
        Map index= getIndex();
        List bundles= getStorageElements(index);

        Set all= new HashSet(bundles.size());
        for (Iterator i= bundles.iterator(); i.hasNext();) {
            Map properties= (Map)i.next();
            STBundle bundle= new STBundle();
            bundle.putAll(properties);
            all.add(bundle);
        }

        return all;
    }


    /**
     * This method translates an arbitrary string into a valid file name.
     * Any '$' or any character that is not a JavaIdentifierPart is replaced by
     * a $ followed by a four digit hex number.
     */
    protected String toValidPath(String string)
    {
        StringBuffer buffer= new StringBuffer();
        int length= string.length();
        for(int i= 0; i < length; i++)
        {
            char c= string.charAt(i);
            if ((Character.isJavaIdentifierPart(c)  ||  Character.isDigit(c))  &&  c != '$')
            {
                buffer.append(c);
            }
            else
            {
                buffer.append('$');
                String hex= Integer.toHexString(c);
                for (int j= hex.length(); j++ < 4;)
                    buffer.append('0');
                buffer.append(hex);
            }
        }
        return buffer.toString();
    }
    protected String fromValidPath(String pathname)
    {
        StringBuffer buffer= new StringBuffer();
        int length= pathname.length();
        for(int i= 0; i < length; i++)
        {
            char c= pathname.charAt(i);
            if (c != '$') {
                buffer.append(c);
            }
            else {
                String hex= pathname.substring(i+1, i+5);
                c= (char)Integer.parseInt(hex, 16);
                buffer.append(c);
                i+= 4;
            }
        }
        return buffer.toString();
    }

    /**
     * Return the jar file for a bundle that has already been installed
     *
     * @return  null if no contents file exists for the given location
     * @throws  BundleException if an error occurs while reading a contents file
     */
    protected STBundleContents getBundleContents(String location)
    throws BundleException
    {
        JarFile jarFile= null;
        try {
            String jarName= getJarName(location);
            File file= new File(jarName);
            if (!file.exists())
                return null;
            jarFile= new JarFile(file);
            STBundleContents contents= new STBundleContents(jarFile);
            return contents;
        }
        catch (IOException x) {
            throw new STException("Error getting jar file for bundle '"+location+"'", x);
        }
        finally {
            if (jarFile != null) {
                try { jarFile.close(); } catch (Throwable t) { }
            }
        }
    }

    /**
     * Read, store, and then return the given bundle's jar file.
     *
     * If an InputStream is specified, the Framework will try to fetch the
     * bundle from the supplied stream.  Otherwise, the Framework will try to
     * download the bundle using its location string, which typically is the
     * string representation of a URL identifying the bundle's codebase.
     *
     * Persist the bundle's contents the bundle's associated data directory.
     */
    STBundleContents installBundle(String source, String location, InputStream in)
    throws BundleException
    {
        /*
        Create the directory in which we'll save the contents file
        */
        String contentDirName= getFolderName(location);
        String jarName= getJarName(location);
        File contentDir= new File(contentDirName);
        boolean createdDir= false;

        InputStream inContent= in;
        OutputStream out= null;
        JarFile jfile= null;
        try {
            /*
            if in is null then open up a stream using location
            */
            if (inContent == null) {
                try {
                    URL url= new URL(source);
                    inContent= url.openStream();
                }
                /*
                if we were not given a valid URL then try assuming that the given
                location is a system resource
                */
                catch (MalformedURLException x) {
                    inContent= ClassLoader.getSystemResourceAsStream(source);
                }
                inContent= new BufferedInputStream(inContent);
            }

            /*
            Create the directory in which we'll save the contents file.
            */
            if (!contentDir.exists())
                if (!contentDir.mkdirs()) {
                    throw new STException("Could not create bundle content folder '"+contentDir+"'");
                }
                else
                    createdDir= true;

            /*
            Open a file to save the contents to
            Read all bytes from the input stream and save to the output stream
            */
            out= new FileOutputStream(jarName);
            out= new BufferedOutputStream(out);
            byte[] buffer= new byte[32 * 1024];
            for (;;) {
                int c= inContent.read(buffer);
                if (c == -1)
                    break;
                out.write(buffer, 0, c);
            }
            out.flush();

            /*
            Finally, read the jar file back in...
            */
            jfile= new JarFile(jarName);
            STBundleContents contents= new STBundleContents(jfile);
            return contents;
        }
        catch (IOException x) {
            /*
            if we created a directory then remove it
            */
            if (createdDir) {
                try { IOUtils.removeAll(contentDirName); } catch (Throwable t) { }
            }

            throw new STException("Error while loading bundle '"+location+"'", x);
        }
        finally {
            if (in == null  &&  inContent != null) {
                try { inContent.close(); } catch (Throwable t) { }
            }
            if (out != null) {
                try { out.close(); } catch (Throwable t) { }
            }
            if (jfile != null) {
                try { jfile.close(); } catch (Throwable t) { }
            }
        }
    }

    /**
     * Back up a bundle's contents file.
     */
    void backupBundle(String location)
    throws IOException
    {
        String jarName= getJarName(location);
        String backupName= jarName+".bak";

        InputStream in= null;
        OutputStream out= null;
        try {

            in = new FileInputStream(jarName);
            in= new BufferedInputStream(in);
            out= new FileOutputStream(backupName);
            out= new BufferedOutputStream(out);
            byte[] buf= new byte[32*1024];
            for (;;) {
                int count= in.read(buf);
                if (count == -1)
                    break;
                out.write(buf, 0, count);
            }
        }
        finally {
            if (in != null) {
                try { in.close(); } catch (Throwable t) { }
            }
            if (out != null) {
                try { out.close(); } catch (Throwable t) { }
            }
        }
    }

    /**
     * Restore a bundle's contents file from the bundle's backup file
     */
    public void restoreBundle(String location)
    throws IOException
    {
        String jarName= getJarName(location);
        String backupName= jarName+".bak";

        InputStream in= null;
        OutputStream out= null;
        try {
            in = new FileInputStream(backupName);
            in= new BufferedInputStream(in);
            out= new FileOutputStream(jarName);
            out= new BufferedOutputStream(out);
            byte[] buf= new byte[32*1024];
            for (;;) {
                int count= in.read(buf);
                if (count == -1)
                    break;
                out.write(buf, 0, count);
            }
        }
        finally {
            if (in != null) {
                try { in.close(); } catch (Throwable t) { }
            }
            if (out != null) {
                try { out.close(); } catch (Throwable t) { }
            }
        }
    }

    /**
     * Remove all traces of the denoted bundle from storage
     */
    void uninstallBundle(String location)
    throws IOException, SAXException
    {
        /*
        Remove the directory associated with the bundle
        */
        String contentDirName= getFolderName(location);
        IOUtils.removeAll(contentDirName);

        /*
        Remove the bundle from the index
        */
        ElementMap index= getIndex();
        List bundles= getStorageElements(index);
        for (int i= bundles.size(); 0 < i--;) {
            Map m= (Map)bundles.get(i);
            String l= LangUtils.toString(m.get(STBundle.KEY_LOCATION));
            if (l != null  &&  l.equals(location))
                bundles.remove(i);
        }
        putBundleElements(index, bundles);
        saveIndex(index);
    }

    String getFolderName(String location) {
        File root= getRootDirectory();
        String rootName= root.getAbsolutePath();
        String contentName= toValidPath(location);
        String contentDirName= rootName+FILE_SEPARATOR+contentName;
        return contentDirName;
    }
    String getJarName(String location) {
        File root= getRootDirectory();
        String rootName= root.getAbsolutePath();
        String contentName= toValidPath(location);
        String contentDirName= rootName+FILE_SEPARATOR+contentName;
        String jarName= contentDirName+FILE_SEPARATOR+"service.tango.bundle.jar";
        return jarName;
    }

    public File getDataFile(String location, String name) {
        String dataName= getFolderName(location)+FILE_SEPARATOR+name;
        return new File(dataName);
    }
}
