/*
   Log.java

   Copyright (c) 2000 by ted stockwell
   All Rights Reserved
*/
package org.servicetango.log;

import org.osgi.framework.*;
import org.osgi.service.log.*;

import java.io.*;
import java.text.*;
import java.util.*;

/**
*   A basic reusable log facility.
*
*   This class implements the "Strategy" pattern whereby it takes an
*   object that does the physical log writing as an argument in its
*   constructor.  Therefore this base class is very reusable.
*
*   This class also has a constructor that creates a logger that writes
*   to a file.
*
*   @author   ted stockwell
*/
public class Log
implements LogService
{
    static final public String KEY_LOG_FILE= "log.file";
    static final public String KEY_LOG_ENABLE= "log.enable";

    /**
     * OSGi Log Types
     */
     static final String[] OSGI_LOG_TYPE= { "all", "error", "warning", "info", "debug" };

     static final String LINE_SEPARATOR= System.getProperty("line.separator");
     static final String FILE_SEPARATOR= System.getProperty("line.separator");
     static final String DEFAULT_ENABLED_TYPES= "error,warning,info,debug";

    private HashSet m_enabled= new HashSet();
    ILogWriter m_writer;

    /**
    *   This constructor creates a Log object that writes
    *   asynchronously to a file-based log.
    *
    *   @param  fileName   denotes the name of the log file.
    *   @param  enabledTypes  a delimited string that denotes what
    *                         log types are enabled.
    */
    public Log(String fileName, String enabledTypes)
    {
        m_writer=
            new AsyncLogWriter(
                new FileLogWriter(fileName));

        init(enabledTypes);
    }


    /**
    *   This constructor creates a Log object that writes
    *   to the given log writer.
    *   The given Properties object should have a Property corresponding
    *   to ServerProperties.KEY_LOG_ENABLE that denotes what log types
    *   are enabled.
    *
    *   @param  enabledTypes  a delimited string that denotes what
    *                         log types are enabled.
    */
    public Log(ILogWriter writer, String enabledTypes)
    {
        m_writer= writer;
        init(enabledTypes);
    }
    public Log(ILogWriter writer) {
        this(writer, DEFAULT_ENABLED_TYPES);
    }

    /**
    *   This constructor creates a Log object that writes
    *   asynchronously to a file-based log.
    *   The given Properties object should have a Property corresponding
    *   to KEY_LOG_FILE that denotes the name of the
    *   log file.
    *   The given Properties object should have a Property corresponding
    *   to KEY_LOG_ENABLE that denotes what log types
    *   are enabled.
    */
    public Log(Map initProperties)
    {
        String logFileName = null;
        if (initProperties != null) {
            Object value= initProperties.get(KEY_LOG_FILE);
            if (value != null)
                logFileName= value.toString();
        }
        m_writer=
            new AsyncLogWriter(
                new FileLogWriter(logFileName));

        init((String)initProperties.get(KEY_LOG_ENABLE));
    }


    private void init(String logTypes)
    {
        if (logTypes != null)
        {
            StringTokenizer tokenizer=
                new StringTokenizer(
                    logTypes,
                    ";:,"+System.getProperty("path.separator"));
            for (; tokenizer.hasMoreTokens();)
            {
                String token= tokenizer.nextToken();
                enabledType(token);
            }
        }
    }


    /**
    *	Disables logging for the given message type
    */
    public void disabledType(String logType)
    {
        m_enabled.remove(logType);
    }
    /**
    *	Disables logging for the given message types
    */
    public void disabledType(String[] logType)
    {
        for (int i= logType.length; 0 < i--;)
            m_enabled.remove(logType[i]);
    }

    /**
    *	Enables logging for the given message type
    */
    public void enabledType(String logType)
    {
        m_enabled.add(logType);
    }
    /**
    *	Enables logging for the given message types
    */
    public void enabledType(String[] logType)
    {
        for (int i= logType.length; 0 < i--;)
            m_enabled.add(logType[i]);
    }

    /**
    *	Returns true if logging Enables logging for the given message
    *	type
    */
    public boolean isEnabledType(String logType)
    {
        return m_enabled.contains(logType);
    }
    /**
    *	Returns true if logging Enables logging for ALL of the given
    *	message types
    */
    public boolean isEnabledType(String[] logType)
    {
        for (int i= logType.length; 0 < i--;)
            if (!m_enabled.contains(logType))
            return false;
        return true;
    }

    /**
    * Log the specified message to the log file (if this Log
    * was created with a valid log file name).
    *
    * @param msg  message to log
    */
    public void log(String[] logType, String msg)
    {
        for (int i= logType.length; 0 <i--;)
        {
            if (isEnabledType(logType[i]))
            {
                m_writer.write(msg);
                break;
            }
        }
    }
    public void log(String logType, String msg)
    {
        if (isEnabledType(logType))
            m_writer.write(msg);
    }
    public void log(String msg)
    {
        m_writer.write(msg);
    }

    public void log(Throwable e, String msg)
    {
        StringWriter writer= new StringWriter();
        e.printStackTrace(new PrintWriter(writer));
        log(msg + LINE_SEPARATOR + writer.toString());
    }
    public void log(String logType, Throwable exception, String msg)
    {
        if (isEnabledType(logType))
            log(exception, msg);
    }
    public void log(String[] logType, Throwable exception, String msg)
    {
        for (int i= logType.length; 0 <i--;)
        {
            if (isEnabledType(logType[i]))
            {
                log(exception, msg);
                break;
            }
        }
    }


    /**
     * Implementation of org.osgi.log.LogService.log method.
     * @see org.osgi.log.LogService.log
     */
    final public void log(ServiceReference service, int level, String message) {
        if (isEnabledType(OSGI_LOG_TYPE[level]))
            log(service, message);
    }
    /**
     * Implementation of org.osgi.log.LogService.log method.
     * @see org.osgi.log.LogService.log
     */
    final public void log(ServiceReference service, int level, String message, Throwable xception) {
        if (isEnabledType(OSGI_LOG_TYPE[level]))
            log(service, message, xception);
    }
    /**
     * Implementation of org.osgi.log.LogService.log method.
     * @see org.osgi.log.LogService.log
     */
    final public void log(int level, String msg, Throwable exception) {
        log(OSGI_LOG_TYPE[level], exception, msg);
    }
    /**
     * Implementation of org.osgi.log.LogService.log method.
     * @see org.osgi.log.LogService.log
     */
    final public void log(int level, String msg) {
        log(OSGI_LOG_TYPE[level], msg);
    }
    public void log(ServiceReference service, String message, Throwable exception) {
        StringWriter writer= new StringWriter();
        exception.printStackTrace(new PrintWriter(writer));
        log(service, message + LINE_SEPARATOR + writer.toString());
    }
    public void log(ServiceReference service, String message) {
        log("["+service.toString()+"]" + LINE_SEPARATOR + message);
    }

    /**
    *   Clear the log file.
    */
    public void clear()
    {
        m_writer.clear();
    }

    /**
    *   Doesn't return until all log messages have been written or a
    *   failure occurs.
    */
    public void flush()
    {
        m_writer.flush();
    }
}
