package org.servicetango.startup;

import org.servicetango.core.*;
import org.osgi.framework.*;
import org.osgi.service.log.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import java.io.*;
import java.util.*;

/**
 * Starter for ServiceTango.
 * Opens a console window that can be used to manage the framework.
 * Configuraed by XML configuration file, servicetango.xml
 *
 * @author  ted stockwell
 */
public class ServiceTango
extends BundleManager
implements Runnable
{
    // Server identification.
    public static final String SERVER_NAME= "ServiceTango";
    public static final String SERVER_VERSION= "v1.0";
    public static final String SERVER_URL= "http://www.servicetango.org";
    public static final String SERVER_INFO= SERVER_NAME+" "+SERVER_VERSION+", "+SERVER_URL;
    public final static String COPYRIGHT= SERVER_NAME + " " + SERVER_VERSION + "\nCopyright (C) 2000 by ted stockwell\nAll Rights Reserved\n";

    public final static String OPTION_PROPERTIES= "-config";
    public final static String OPTION_HELP= "-?";

    /**
    * default properties file name
    */
    public final static String DEFAULT_PROPERTIES_FILE= "servicetango.xml";

    /**
    *  Main routine for running this server from the command line.
    */
    public static void main( String[] args ) throws Exception
    {
        try
        {
            String propertiesFile= DEFAULT_PROPERTIES_FILE;

            // Parse args.
            for (int i= 0; i < args.length; i++) {
                if (args[i].equalsIgnoreCase(OPTION_PROPERTIES)) {
                    if (++i < args.length) {
                        propertiesFile= args[i];
                    }
                    else
                        throw new Exception("Missing value for " + OPTION_PROPERTIES + " option");
                }
                if (args[i].equalsIgnoreCase(OPTION_HELP)) {
                    usage(System.out);
                    System.exit(0);
                }
            }

            try {
                // Create and start the server.
                m_instance= new ServiceTango(new File(propertiesFile));
            }
            catch (FileNotFoundException e) {
                throw new FileNotFoundException("Could not find the server's configuration file '"+propertiesFile+"'.");
            }
            catch (IOException e) {
                throw new IOException("Error reading the server's configuration file '"+propertiesFile+"';\n" + e);
            }
            catch (SAXParseException e) {
                throw new IOException("The configuration file '"+propertiesFile+"' is not a valid XML file;\n" + e);
            }

            m_instance.run();
        }
        catch (Exception e) {
            e.printStackTrace();
            usage();
            System.exit(1);
        }
        finally {
        }
    }

    private static void usage() {
        usage(System.err);
    }
    private static void usage(PrintStream out) {
        out.println("ServiceTango");
        out.println("USAGE:");
        out.println("java com.servicetango.startup.ServiceTango [options]");
        out.println("where [options] may include the following:");
        out.println("-?  Prints this usage message.");
        out.println("-properties <properties>  The URL of a file that contains servlet properties.");
        out.println("If not specified then \"servicetango.xml\" is used");
        out.flush();
    }


    protected STConfiguration m_properties;
    File m_propertiesFile= null;
    long m_propertiesLastModified= 0;
    boolean m_halted= false;
    BufferedReader in= new BufferedReader(new InputStreamReader(System.in));
    PrintStream out= System.out;

    /**
     * A reference to the main server object
     */
    static private ServiceTango m_instance;

    /**
    * @param  Map   server properties
    * @param  port         server port
    */
    private ServiceTango(STConfiguration properties)
    throws BundleException
    {
        super(properties);
        m_properties= properties;

        init(false/*don't reload properties file*/);
    }

    /**
    * @param  Map   server properties
    * @param  port         server port
    */
    public ServiceTango(File propertiesFile)
    throws FileNotFoundException, IOException, BundleException, SAXException
    {
        this(ServiceTango.loadPropertiesFile(propertiesFile));
        m_propertiesFile = propertiesFile;
    }

    synchronized public void startUp()
    throws BundleException
    {
        init(true);
        run();
    }

    synchronized public void shutDown() {
        super.shutDown();
        halt();
    }

    synchronized void halt() {
        m_halted= true;
    }

    synchronized private void init(boolean reload)
    throws BundleException
    {
        /*
        start all bundles
        */
        super.startUp();

        if (m_propertiesFile == null)
            return;

        /*
        Load the properties.
        Properties were successfully loaded when the server was started.
        If Properties cannot be loaded now then just log the error and
        continue using the current set of properties.
        */
        m_propertiesLastModified= m_propertiesFile.lastModified();
        if (reload  ||  m_properties == null)
        {
            try {
                m_properties= loadPropertiesFile(m_propertiesFile);

                // call CgiServletContext.init to reset properties
                init(m_properties);
            }
            catch (Throwable t) {
                getLogService().log(LogService.LOG_ERROR, "Could not reload properties file " + m_propertiesFile, t);
            }
        }
    }



    /**
     *  Load a properties file
     */
    static protected STConfiguration loadPropertiesFile(File propertiesFile)
    throws FileNotFoundException, IOException, SAXException
    {
        // load the file
        InputStream propertiesInput= new FileInputStream(propertiesFile);
        try {
            propertiesInput= new BufferedInputStream(propertiesInput);
            STConfiguration properties= new STConfiguration(propertiesInput);
            return properties;
        }
        finally {
            try {
                propertiesInput.close();
            }
            catch (IOException e) {
            }
        }
    }

    interface Command {
        public String help();
        public String description();
        public void run(String[] args) throws Exception;
    }
    Command getCommand= new Command() {
        public void run(String[] args) {
            STConfiguration config= getConfiguration();
            if (args.length < 2) {
                String[] keys= config.getSTPropertyKeys();
                for (int i= 0; i < keys.length; i++) {
                    Object value= config.getSTProperty(keys[i]);
                    out.println(keys[i]+"="+value);
                }
                return;
            }

            for (int i= 1; i < args.length; i++) {
                String key= args[i];
                String value= config.getSTProperty(key);
                if (value == null) {
                    out.println("'"+key+"' is not recognised");
                }
                else
                    out.println(key+"="+value);
            }
        }
        public String description() {
            return "Get a configuration property";
        }
        public String help() {
            return "g[et] [<property name>]";
        }
    };
    Command setCommand= new Command() {
        public void run(String[] args) {
            if (args.length < 2) {
                out.println(help());
                return;
            }

            StringTokenizer tokenizer= new StringTokenizer(args[1], "=");
            List l= new ArrayList();
            for (; tokenizer.hasMoreTokens();)
                l.add(tokenizer.nextToken());
            if (l.size() < 2) {
                out.println(help());
                return;
            }
            args= new String[l.size()];
            l.toArray(args);
            getConfiguration().setSTProperty(args[0], args[1]);
        }
        public String description() {
            return "Set a configuration property";
        }
        public String help() {
            return "s[et] <property-name>=<property-value>";
        }
    };
    Map statusText= new HashMap();
    {
        statusText.put(new Integer(Bundle.ACTIVE), "ACTIVE");
        statusText.put(new Integer(Bundle.INSTALLED), "INSTALLED");
        statusText.put(new Integer(Bundle.RESOLVED), "RESOLVED");
        statusText.put(new Integer(Bundle.STARTING), "STARTING");
        statusText.put(new Integer(Bundle.STOPPING), "STOPPING");
        statusText.put(new Integer(Bundle.UNINSTALLED), "UNINSTALLED");
    }
    Command bundleCommand= new Command() {
        public void run(String[] args) {
            STBundle[] bundles= getBundles();
            if (bundles.length <= 0) {
                out.println("No bundles are installed");
                return;
            }
            out.println("  Id |      Status | Location");
            for (int i= 0; i < bundles.length; i++) {
                STBundle bundle= bundles[i];
                String id= "    "+bundle.getBundleId();
                id= id.substring(id.length() - 4);
                String status= "           "+statusText.get(new Integer(bundle.getState()));
                status= status.substring(status.length() -11);
                out.println(id+" | "+status+" | "+bundle.getLocation());
            }
        }
        public String description() {
            return "List all installed bundles";
        }
        public String help() {
            return "b[undles]";
        }
    };
    Command servicesCommand= new Command() {
        public void run(String[] args) {
            ServiceReference[] references;
            if (args.length < 2) {
                references= getRegisteredServices();
            }
            else {
                try {
                    references= getServiceReferences(null, args[1]);
                }
                catch (InvalidSyntaxException x) {
                    out.println("Not a valid filter");
                    return;
                }
            }
            for (int i= 0; i < references.length; i++) {
                if (0 < i)
                    out.println("---------------------------");
                STServiceReference ref= (STServiceReference)references[i];

                String[] clazzes= ref.getClasses();
                if (clazzes.length == 1) {
                    out.println("Service Class:"+clazzes[0]);
                }
                else {
                    out.println("Service Classes:");
                    for (int j= 0; j < clazzes.length; j++)
                        out.println("  "+clazzes[j]);
                }

                Bundle bundle= ref.getBundle();
                out.println("Registered By Bundle "+bundle.getBundleId()+"|"+bundle.getLocation());

                Bundle[] bundles= getBundles();
                boolean usesOtherBundles= false;
                for (int j= 0; j < bundles.length; j++) {
                    Bundle b= bundles[j];
                    long useCount= ref.getUseCount(b.getBundleId());
                    if (0 < useCount) {
                        if (!usesOtherBundles) {
                            out.println();
                            out.println("Use Count | Bundle Id | Bundle Location");
                        }
                        usesOtherBundles= true;

                        String id= "         "+bundle.getBundleId();
                        id= id.substring(id.length() - 9);
                        String sUseCount= "         "+useCount;
                        sUseCount= sUseCount.substring(sUseCount.length() - 9);
                        out.println(sUseCount+" | "+id+" | "+b.getLocation());
                    }
                }
                if (!usesOtherBundles)
                    out.println("This service is not using any other bundles.");

                String[] keys= ref.getPropertyKeys();
                out.println("Properties:");
                for (int j= 0; j < keys.length; j++) {
                    String key= keys[j];
                    Object property= ref.getProperty(key);
                    if (property instanceof Object[]) {
                        Object[] a= (Object[])property;
                        out.print("  "+key+"=");
                        for (int k= 0; k < a.length; k++) {
                            if (0 < k)
                                out.print(",");
                            out.print(a[k]);
                        }
                        out.println();
                    }
                    else {
                        String value= property.toString();
                        out.println("  "+key+"="+value);
                    }
                }
            }
        }
        public String description() {
            return "List all available services that match a given LDAP search filter."
                    +"\nIf no filter is specified then all available services are listed.";
        }
        public String help() {
            return "ser[vices] [<filter>]";
        }
    };
    Command installCommand= new Command() {
        public void run(String[] args)
        throws BundleException
        {
            if (args.length < 2) {
                out.println(help());
                return;
            }
            installBundle(args[1]);
            out.println("'"+args[1]+"' successfully installed");
        }
        public String description() {
            return "Install a bundle.";
        }
        public String help() {
            return "i[nstall] <bundle-url>..."
                   +"\n   example: install file:///c:/jes-2_0-ea1/bundles/log.jar";
        }
    };
    Command startCommand= new Command() {
        public void run(String[] args) {
            if (args.length < 2) {
                out.println(help());
                return;
            }
            for (int i= 1; i < args.length; i++) {
                long id;
                STBundle bundle;
                try {
                    id= Long.parseLong(args[1]);
                    bundle= getBundle(id);
                }
                /*
                if not a number then assume the user type a bundle location
                */
                catch (NumberFormatException x) {
                    bundle= getBundle(args[1]);
                    if (bundle == null) {
                        out.println("Unrecognized bundle location '"+args[i]+"'");
                        continue;
                    }
                }
                try {
                    startBundle(bundle.getBundleId());
                    out.println("Bundle "+args[i]+" successfully started");
                }
                catch (Exception x) {
                    out.println("Error starting bundle '"+args[i]+"'");
                    x.printStackTrace(out);
                }
            }
        }
        public String description() {
            return "Start a bundle given the bundle's original URL or the bundle's id.";
        }
        public String help() {
            return "sta[rt] <bundle-url|bundle-id>...";
        }
    };
    Command stopCommand= new Command() {
        public void run(String[] args) {
            if (args.length < 2) {
                out.println(help());
                return;
            }
            for (int i= 1; i < args.length; i++) {
                STBundle bundle;
                try {
                    long id= Long.parseLong(args[1]);
                    bundle= getBundle(id);
                }
                /*
                if not a number then assume the user typed a bundle location
                */
                catch (NumberFormatException x) {
                    bundle= getBundle(args[1]);
                    if (bundle == null) {
                        out.println("Unrecognized bundle location '"+args[i]+"'");
                        continue;
                    }
                }
                try {
                    stopBundle(bundle.getBundleId());
                    out.println("Bundle "+args[i]+" successfully stopped");
                }
                catch (Exception x) {
                    out.println("Error stopping bundle '"+args[i]+"'");
                    x.printStackTrace(out);
                }
            }
        }
        public String description() {
            return "Stop a bundle given the bundle's original URL or the bundle's id.";
        }
        public String help() {
            return "sto[p] <bundle-url|bundle-id>...";
        }
    };
    Command uninstallCommand= new Command() {
        public void run(String[] args) {
            if (args.length < 2) {
                out.println(help());
                return;
            }
            for (int i= 1; i < args.length; i++) {
                STBundle bundle;
                try {
                    long id= Long.parseLong(args[1]);
                    bundle= getBundle(id);
                }
                /*
                if not a number then assume the user typed a bundle location
                */
                catch (NumberFormatException x) {
                    bundle= getBundle(args[1]);
                    if (bundle == null) {
                        out.println("Unrecognized bundle location '"+args[i]+"'");
                        continue;
                    }
                }
                try {
                    uninstallBundle(bundle.getBundleId());
                    out.println("Bundle "+args[i]+" successfully uninstalled");
                }
                catch (Exception x) {
                    out.println("Error unistalling bundle '"+args[i]+"'");
                    x.printStackTrace(out);
                }
            }
        }
        public String description() {
            return "Uninstall a bundle given the bundle's original URL or the bundle's id.";
        }
        public String help() {
            return "un[install] <bundle-url|bundle-id>...";
        }
    };
    Command updateCommand= new Command() {
        public void run(String[] args) {
            if (args.length < 2) {
                out.println(help());
                return;
            }
            for (int i= 1; i < args.length; i++) {
                STBundle bundle;
                try {
                    long id= Long.parseLong(args[1]);
                    bundle= getBundle(id);
                }
                /*
                if not a number then assume the user typed a bundle location
                */
                catch (NumberFormatException x) {
                    bundle= getBundle(args[1]);
                    if (bundle == null) {
                        out.println("Unrecognized bundle location '"+args[i]+"'");
                        continue;
                    }
                }
                try {
                    updateBundle(bundle.getBundleId());
                    out.println("Bundle "+args[i]+" successfully updated");
                }
                catch (Exception x) {
                    out.println("Error updating bundle '"+args[i]+"'");
                    x.printStackTrace(out);
                }
            }
        }
        public String description() {
            return "Update a bundle given the bundle's original URL or the bundle's id.";
        }
        public String help() {
            return "up[date] <bundle-id> [<bundle-url]";
        }
    };
    Command shutdownCommand= new Command() {
        public void run(String[] args) {
            shutDown(); // shut down bundle manager
        }
        public String description() {
            return "Shutdown the gateway.";
        }
        public String help() {
            return "sh[utdown]";
        }
    };
    Command helpCommand= new Command() {
        public void run(String[] args) {
            if (1 < args.length) {
                Command c= (Command)commands.get(args[1]);
                if (c == null) {
                    out.println("unknown command '"+args[1]+"'");
                    return;
                }
                out.println(c.description());
                out.println(c.help());
                return;
            }

            HashSet set= new HashSet(commands.values());
            for (Iterator i= set.iterator(); i.hasNext();)
                out.println(((Command)i.next()).help());
            out.println("\n  <token>... is a comma separated list of <token>s\n");
        }
        public String description() {
            return "Get help on a given command"
                +"\nIf no command is specified the list a summary of all commands.";
        }
        public String help() {
            return "h[elp] [<command>]";
        }
    };

    Map commands= new HashMap();
    {
        commands.put("get", getCommand);
        commands.put("g", getCommand);
        commands.put("set", setCommand);
        commands.put("s", setCommand);
        commands.put("bundle", bundleCommand);
        commands.put("b", bundleCommand);
        commands.put("services", servicesCommand);
        commands.put("ser", servicesCommand);
        commands.put("install", installCommand);
        commands.put("i", installCommand);
        commands.put("start", startCommand);
        commands.put("sta", startCommand);
        commands.put("stop", stopCommand);
        commands.put("sto", stopCommand);
        commands.put("uninstall", uninstallCommand);
        commands.put("un", uninstallCommand);
        commands.put("update", updateCommand);
        commands.put("up", updateCommand);
        commands.put("shutdown", shutdownCommand);
        commands.put("sh", shutdownCommand);
        commands.put("help", helpCommand);
        commands.put("h", helpCommand);
        commands.put("?", helpCommand);
    }

    public void run() {
        out.println(COPYRIGHT);
        try {
            bundleCommand.run(new String[] { "b" });
            out.println();
            for (;!m_halted;) {
                out.print(SERVER_NAME+"> ");
                String readLine= in.readLine();
                StringTokenizer tokenizer= new StringTokenizer(readLine, " \t");
                List l= new ArrayList();
                for (; tokenizer.hasMoreTokens();)
                    l.add(tokenizer.nextToken());
                if (l.size() <= 0)
                    continue;
                String[] args= new String[l.size()];
                l.toArray(args);
                Command c= (Command)commands.get(args[0]);
                if (c == null) {
                    out.println("Invalid command:"+args[0]);
                    continue;
                }
                try {
                    c.run(args);
                }
                catch (Exception x) {
                    out.print("Internal Error: ");
                    x.printStackTrace(out);
                }
            }
        }
        catch (Exception x) {
            x.printStackTrace(out);
        }

        shutDown();
    }
}



