/*
   DateUtils.java

   Copyright (c) 2000 by ted stockwell
   All Rights Reserved
*/
package org.servicetango.util;

import java.lang.ref.SoftReference;
import java.text.*;
import java.util.*;

/**
 * Miscellaneous Date utilities
 *
 * @author ted stockwell [emorning@mediaone.net]
 */
public class DateUtils
{
    public static final Date EPOCH_DATE= new Date(0);

    private static SoftReference gmtFormatter = null;

    private static final int MONTH_LENGTH[]
        = {31,28,31,30,31,30,31,31,30,31,30,31}; // 0-based

    protected static final SimpleDateFormat[] m_dateFormats=
    {
        new SimpleDateFormat("MM/dd/yyyy hh:mm:ss"),
        new SimpleDateFormat("MM/dd/yyyy hh:mm:ss")
    };

    private static final ArrayList gCalendarPool= new ArrayList();

    static public final DateFormat MONTH_FORMAT= new SimpleDateFormat("MMMMMMMMMMMMM");

    /**
     * Returns a calendar from an internal pool.
     * The caller may only set the time property of the returned Calendar (via
     * the setTime method).
     * If any other properties are set by the caller then the returned Calendar
     * should not be returned to the pool.
     *
     * This method uses the Calendar.getInstance() method to create Calendars
     */
    static public Calendar getPoolCalendar() {
        Calendar c= null;
        synchronized (gCalendarPool) {
            if (0 < gCalendarPool.size())
                c= (Calendar)gCalendarPool.remove(0);
        }
        if (c == null)
            c= Calendar.getInstance();
        c.setTime(EPOCH_DATE);
        return c;
    }
    static public final Calendar getPoolCalendar(Date d) {
        Calendar c= getPoolCalendar();
        c.setTime(d);
        return c;
    }
    static public final Calendar getEpochCalendar() {
        Calendar c= getPoolCalendar();
        c.setTime(EPOCH_DATE);
        return c;
    }
    static public void poolCalendar(Calendar calendar) {
        synchronized (gCalendarPool) {
            gCalendarPool.add(calendar);
        }
    }

    /**
     * Returns a DateFormat that formats a date relative to the current time.
     * The closer the given Date is to the current Time the more abbreviated
     * the returned DateFormat is.
     * If the given Date is on the same day as the current Date then no day is
     * included in the returned format.
     * If the given Date is in the same month as the current Date then no
     * month is included in the returned format.
     * If the given Date is in the same year as the current Date then no
     * year is included in the returned format.
     *
     *  !!! UNFINISHED TODO
     */
    static public DateFormat getRelativeFormat(Date date)
    {
        return DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT);
    }

    private static final String[] weekdays =
	{ "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
    public static final String months[] = { "Jan", "Feb", "Mar", "Apr",
                                          "May", "Jun", "Jul", "Aug",
                                          "Sep", "Oct", "Nov", "Dec" };

    /**
     * Converts a Date into an RFC-1123 string.
     */
    public static String toRfc1123String(Date date)
	{
	// We have to go through some machinations here to get the
	// correct day of the week in GMT.  getDay() gives the day in
	// local time.  getDate() gives the day of the month in local
	// time.  toGMTString() gives a formatted string in GMT.  So, we
	// extract the day of the month from the GMT string, and if it
	// doesn't match the local one we change the local day of the
	// week accordingly.
	//
	// The Date class sucks.
        Calendar calendar= getPoolCalendar();
        int localDay = calendar.get(Calendar.DAY_OF_WEEK);
	int localDate = calendar.get(Calendar.DAY_OF_MONTH);
        poolCalendar(calendar);
	String gmtStr = DateUtils.toGMTString(date);
	int blank = gmtStr.indexOf( ' ' );
	int gmtDate = Integer.parseInt( gmtStr.substring( 0, blank ) );
	int gmtDay;
	if ( gmtDate > localDate || ( gmtDate < localDate && gmtDate == 1 ) )
	    gmtDay = ( localDay + 1 ) % 7;
	else if ( localDate > gmtDate || ( localDate < gmtDate && localDate == 1 ) )
	    gmtDay = ( localDay + 6 ) % 7;
	else
	    gmtDay = localDay;
	return weekdays[gmtDay] + ( gmtDate < 10 ? ", 0" : ", " ) + gmtStr;
    }

  /**
   * ISO 8601 and http://www.w3.org/TR/NOTE-datetime. YYYY-MM-DDThh:mm:ss.
   * where the 'T' is a literal 'T'.
   */
  public static String toISO8601String(Date date) {
    // FIXME: include timezone info?
    Calendar c= Calendar.getInstance();
    c.setTime(date);
    return "" + (c.get(Calendar.YEAR)) + '-' +
           twoDigits(c.get(Calendar.MONTH) + 1) + '-' +
           twoDigits(c.get(Calendar.DAY_OF_MONTH)) + 'T' +
           time(date);
  }

  /**
   * Returns date in RFC 850 (obsoleted by RFC 1036) format.
   * Used eg by Netscape's Cookie proposal.
   * The format is like: "Sunday, 06-Nov-94 08:49:37 GMT".
   */
  public static String toRfc850String(Date date) {
    Calendar c= Calendar.getInstance();
    c.setTime(date);
    return weekdays[c.get(Calendar.DAY_OF_WEEK)] + ", " +
           twoDigits(c.get(Calendar.DAY_OF_MONTH)) + '-' +
           months[c.get(Calendar.MONTH)] + '-' +
           twoDigits(c.get(Calendar.YEAR)) + ' ' +
           time(date) + " GMT";
  }

  /**
   * Returns time as "08:49:37" - common for several time formats.
   */
  public static String time(Date date) {
    Calendar c= Calendar.getInstance();
    return twoDigits(c.get(Calendar.HOUR)) + ':' +
           twoDigits(c.get(Calendar.MINUTE)) + ':' +
           twoDigits(c.get(Calendar.SECOND));
  }

  /**
   * Returns the given int as String with at least two digits, e.g.
   * twoDigits(7) -> "07" and twoDigits(42) -> "42".
   */
  public static String twoDigits(int i) {
    String s = Integer.toString(i);
    if (s.length() > 1) return s;
    else return "0" + s;
  }

    /**
    *   Returns the # of days in a month.
    *   Months are 0-based, starting in January
    */
    public static final int getDaysInMonth(int month) {
        return MONTH_LENGTH[month];
    }

    /**
    *   Returns the # of days in a month associated with the given Date.
    *   Months are 0-based, starting in January
    */
    public static final int getDaysInMonth(Calendar c) {
        int month= c.get(Calendar.MONTH);
        return MONTH_LENGTH[month];
    }
    public static int getDaysInMonth(Date date) {
        Calendar c= getPoolCalendar();
        c.setTime(date);
        int daysInMonth= getDaysInMonth(c);
        poolCalendar(c);
        return daysInMonth;
    }

    /**
    *   Returns the # of weeks in the month associated with the given
    *   Date.  Returns 4, 5, or 6.
    */
    public static int getWeeksInMonth(Calendar c) {
        int daysInMonth= MONTH_LENGTH[c.get(Calendar.MONTH)];
        int weeksInMonth= 5;
        c.set(Calendar.DAY_OF_MONTH, 1);
        int dayOfWeek= c.get(Calendar.DAY_OF_WEEK);

        switch (daysInMonth) {
            case 28:
                if (dayOfWeek == 1) weeksInMonth= 4; break;
            case 30:
                if (dayOfWeek == 7) weeksInMonth= 6; break;
            case 31:
                if (6 <= dayOfWeek) weeksInMonth= 6; break;
        }
        return weeksInMonth;
    }
    public static int getWeeksInMonth(Date date) {
        Calendar c= getPoolCalendar();
        c.setTime(date);
        int weeksInMonth= getWeeksInMonth(c);
        poolCalendar(c);
        return weeksInMonth;
    }

    /**
    *   Returns the week of the month associated with the date.
    *   Result is 0-based.
    */
    public static int getWeekOfMonth(Calendar c) {
        int dayOMonth= c.get(Calendar.DAY_OF_MONTH); // between 1 and 31
        int week= 0;
        c.set(Calendar.DAY_OF_MONTH, 1);
        int daysInFirstWeek= 7 - c.get(Calendar.DAY_OF_WEEK);
        if (daysInFirstWeek < dayOMonth) {
            int wholeWeeks= dayOMonth/7;
            week+= wholeWeeks;
            dayOMonth-= wholeWeeks * 7;
            if (0 < dayOMonth)
                week++;
        }
        return week;
    }
    public static int getWeekOfMonth(Date date) {
        Calendar c= getPoolCalendar();
        c.setTime(date);
        int week= getWeekOfMonth(c);
        poolCalendar(c);
        return week;
    }


    /**
    *   Returns a Date that is exactly count months after the given date.
    *   If count < 0 then the months are decremented.
    */
    public static Date incrMonth(Date oldDate, int count) {
        if (12 <= count || count <= -12) {
            Calendar c= getPoolCalendar(oldDate);
            int years= count/12;
            c.add(Calendar.YEAR, years);
            oldDate= c.getTime();
            count-= years * 12;
            poolCalendar(c);
        }
        Calendar c= getPoolCalendar(oldDate);
        c.add(Calendar.MONTH, count);
        Date date= c.getTime();
        poolCalendar(c);

        return date;
    }
    /**
    *   Returns a Date that is exactly one month after the given date
    */
    public final static Date incrMonth(Date date) {
        return incrMonth(date, 1);
    }

    /**
    *   Returns a Date that is exactly count months before the given date.
    *   If count < 0 then the months are decremented.
    */
    public final static Date decrMonth(Date oldDate, int count) {
        return incrMonth(oldDate, -count);
    }
    /**
    *   Returns a Date that is exactly one month before the given date
    */
    public final static Date decrMonth(Date date) {
        return decrMonth(date, 1);
    }



    /**
    *   Returns a Date that is exactly count days after the given date.
    */
    public static Date incrDate(Date oldDate, int count) {
        if (365 <= count || count <= -365) {
            Calendar c= getPoolCalendar(oldDate);
            int years= count/365;
            c.add(Calendar.YEAR, years);
            oldDate= c.getTime();
            count-= years * 365;
            poolCalendar(c);
        }
        Calendar c= getPoolCalendar(oldDate);
        c.add(Calendar.DATE, count);
        Date date= c.getTime();
        poolCalendar(c);

        return date;
    }
    /**
    *   Returns a Date that is exactly one day after the given date.
    */
    public static Date incrDate(Date oldDate) {
        return incrDate(oldDate, 1);
    }

    /**
    *   Returns a Date that is exactly count days before the given
    *   date.
    */
    public final static Date decrDate(Date oldDate, int count) {
        return incrDate(oldDate, -count);
    }
    /**
    *   Returns a Date that is exactly one day before the given date.
    */
    public static Date decrDate(Date oldDate) {
        return decrDate(oldDate, 1);
    }


    /**
    *   Returns a string representation of the given month (0-based).
    */
    public static final String getMonthName(int month) {
        if (MONTH_NAMES == null) {
            synchronized (DateUtils.class) {
                if (MONTH_NAMES == null) {
                    MONTH_NAMES= new String[12];
                    Calendar c= getPoolCalendar();
                    c.set(Calendar.DAY_OF_MONTH, 1);
                    for (int i= Calendar.JANUARY; i <= Calendar.DECEMBER; i++) {
                        c.set(Calendar.MONTH, i);
                        String name= MONTH_FORMAT.format(c.getTime());
                        MONTH_NAMES[i]= name;
                    }
                    poolCalendar(c);
                }
            }
        }
        return MONTH_NAMES[month];
    }
    private static String MONTH_NAMES[]= null;



    /**
    *   Returns a Date that is exactly count hours after the given date.
    */
    public static Date incrHours(Date oldDate, int count) {
        if (24 <= count) {
            int days= count/24;
            oldDate= incrDate(oldDate, days);
            count-= days * 24;
        }

        Calendar c= getPoolCalendar(oldDate);
        c.add(Calendar.HOUR, count);
        Date date= c.getTime();
        poolCalendar(c);
        return date;
    }
    /**
    *   Returns a Date that is exactly one hour after the given date.
    */
    public static Date incrHour(Date oldDate) {
        return incrHours(oldDate, 1);
    }

    /**
    *   Returns a Date that is exactly count days before the given date.
    */
    public static Date decrHours(Date oldDate, int count) {
        return incrHours(oldDate, -count);
    }
    /**
    *   Returns a Date that is exactly one day before the given date.
    */
    public static Date decrHour(Date oldDate) {
        return decrHours(oldDate, 1);
    }






    /**
    *   Returns a Date that is exactly count minutes after the given date.
    */
    public static Date incrMinutes(Date oldDate, int count) {
        if (60 <= count) {
            int hours= count/60;
            oldDate= incrHours(oldDate, hours);
            count-= hours * 60;
        }

        Calendar c= getPoolCalendar(oldDate);
        c.add(Calendar.MINUTE, count);
        Date date= c.getTime();
        poolCalendar(c);
        return date;
    }
    /**
    *   Returns a Date that is exactly one minute after the given date.
    */
    public static Date incrMinute(Date oldDate) {
        return incrMinutes(oldDate, 1);
    }

    /**
    *   Returns a Date that is exactly count minutes before the given date.
    */
    public static Date decrMinutes(Date oldDate, int count) {
        return incrMinutes(oldDate, -count);
    }
    /**
    *   Returns a Date that is exactly one minute before the given date.
    */
    public static Date decrMinute(Date oldDate) {
        return decrMinutes(oldDate, 1);
    }

    /**
     * Creates a string representation of this <tt>Date</tt> object of
     * the form:
     * <blockquote<pre>
     * d mon yyyy hh:mm:ss GMT</pre></blockquote>
     * where:<ul>
     * <li><i>d</i> is the day of the month (<tt>1</tt> through <tt>31</tt>),
     *     as one or two decimal digits.
     * <li><i>mon</i> is the month (<tt>Jan, Feb, Mar, Apr, May, Jun, Jul,
     *     Aug, Sep, Oct, Nov, Dec</tt>).
     * <li><i>yyyy</i> i sthe year, as four decimal digits.
     * <li><i>hh</i> is the hour of the day (<tt>00</tt> through <tt>23</tt>),
     *     as two decimal digits.
     * <li><i>mm</i> is the minute within the hour (<tt>00</tt> through
     *     <tt>59</tt>), as two decimal digits.
     * <li><i>ss</i> is the second within the minute (<tt>00</tt> through
     *     <tt>61</tt>), as two decimal digits.
     * <li><i>GMT</i> is exactly the ASCII letters "<tt>GMT</tt>" to indicate
     *     Greenwich Mean Time.
     * </ul><p>
     * The result does not depend on the local time zone.
     *
     * @return  a string representation of this date, using the Internet GMT
     *          conventions.
     * @see     java.text.DateFormat
     * @see     java.util.Date#toString()
     * @see     java.util.Date#toLocaleString()
     * @deprecated As of JDK version 1.1,
     * replaced by <code>DateFormat.format(Date date)</code>, using a
     * GMT <code>TimeZone</code>.
     */
    static public String toGMTString(Date date) {
        DateFormat formatter = null;
        if (gmtFormatter != null)
            formatter = (DateFormat)gmtFormatter.get();
        if (formatter == null) {
            /* No cache yet, or cached formatter GC'd */
            formatter = new SimpleDateFormat("d MMM yyyy HH:mm:ss 'GMT'",  Locale.US);
            formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
            gmtFormatter = new SoftReference(formatter);
        }
        return formatter.format(date);
    }

    static public int getCurrentYear() {
        Calendar c= getPoolCalendar(new Date());
        int year= c.get(Calendar.YEAR);
        poolCalendar(c);
        return year;
    }

}
