/*
   ElementMap.java

   Copyright (C) 1999 by GolfTango
   All Rights Reserved
*/

package org.servicetango.util;

import org.servicetango.util.*;

import java.util.*;
import java.io.*;

import org.w3c.dom.*;
import org.xml.sax.*;


import javax.xml.parsers.*;
import com.sun.xml.tree.*;

/**
 * Provides a Map interface facade to an XML Node object.
 * Element Nodes in the given Node are treated as property keys and the text
 * values of those element nodes are treated as property values.
 *
 * If the Map.get method is called and there is more than one element with that
 * name then a List object is returned that contains either a Map object or String
 * for every associated Element.
 *
 * I frequently create "Property Container" objects based on Maps (objects that
 * can have properties added/removed at run-time, see "San Francisco Design
 * Patterns" for an explanation of this pattern).  A Map interface to an XML
 * element allows me to use this Map-based pattern without losing the other
 * elements in an XML element - like comments.
 *
 * @author  ted stockwell [emorning@mediaone.net]
 */
public class ElementMap
extends AbstractMap
implements java.io.Serializable
{
    static final String LINE_SEPARATOR= (String)System.getProperty("line.separator");
    static final String LINE_SEPARATORS_3= LINE_SEPARATOR+LINE_SEPARATOR+LINE_SEPARATOR;
    static final String LINE_SEPARATORS_2= LINE_SEPARATOR+LINE_SEPARATOR;

    Element m_element= null;
    public ElementMap() {
        try {
            InputStream in= new java.io.ByteArrayInputStream(("<empty-node></empty-node>").getBytes());
            DocumentBuilderFactory docBuilderFactory= DocumentBuilderFactory.newInstance();
            DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
            Document docRequest= docBuilder.parse(in);
            m_element= docRequest.getDocumentElement();
        }
        catch (SAXException x) { // not gonna happen
        }
        catch (IOException x) { // not gonna happen
        }
        catch (ParserConfigurationException x) { // not gonna happen
        }
    }
    public ElementMap(Element element) {
        m_element= element;
    }
    public ElementMap(InputStream in)
    throws IOException, SAXException
    {
        load(in);
    }


    NodeList m_nodes= null;
    NodeList getNodeList() {
        if (m_nodes != null)
            return m_nodes;
        m_nodes= XMLUtils.getElementChildNodes(m_element);
        return m_nodes;
    }
    void clearNodeList() {
        m_nodes= null;
    }

    public int size() {
        return getNodeList().getLength();
    }
    public boolean isEmpty() {
        return getNodeList().getLength() <= 0;
    }
    public boolean containsKey(Object key) {
        NodeList nl= getNodeList();
        for (int i= nl.getLength(); 0 < i--;) {
            if (nl.item(i).getNodeName().equals(key))
                return true;
	}
        return false;
    }
    public boolean containsValue(Object value) {
        NodeList nl= getNodeList();
        for (int i= nl.getLength(); 0 < i--;) {
            if (nl.item(i).getNodeValue().equals(value))
                return true;
	}
        return false;
    }
    public Object get(Object key) {
        NodeList nl= getNodeList();
        Object value= null;
        for (int i= nl.getLength(); 0 < i--;) {
            Node n= nl.item(i);
            if (n.getNodeName().equals(key)) {

                /*
                create an object associated with the node.
                if the node contains child element nodes then the element should
                be represented by another ElementMap otherwise the element should
                be represented by a String that contains all the node's text.
                */
                Object nodeObject= null;
                NodeList nlt= XMLUtils.getElementChildNodes(n);
                if (0 < nlt.getLength()) {
                    nodeObject= new ElementMap((Element)n);
                }
                else {
                    nlt= XMLUtils.getTextChildNodes(n);
                    if (0 < nlt.getLength()) {
                        String allText= nlt.item(0).getNodeValue();
                        for (int j= 1; j < nlt.getLength(); j++)
                            allText+= nlt.item(j).getNodeValue();
                        nodeObject= allText;
                    }
                    else
                        nodeObject= "";
                }

                /*
                if this is the first object we've found that matches the key then
                our return value will be just this object.  If we find more objects
                that match the given key then we will return a List that contains all
                the matching objects.
                */
                if (value == null) {
                    value= nodeObject;
                }
                else {
                    if (!(value instanceof List)) {
                        List l= new ArrayList(2);
                        l.add(value);
                        value= l;
                    }
                    ((List)value).add(nodeObject);
                }
            }
	}
        return value;
    }
    public Object put(Object key, Object value) {
        Object old= remove(key);
        clearNodeList();

        Document doc= m_element.getOwnerDocument();
        if (value instanceof List) {
            for (Iterator i= ((List)value).iterator(); i.hasNext();) {
                Object next= i.next();
                Element child= createElement(key, next);
                m_element.appendChild(doc.createTextNode(LINE_SEPARATOR));
                m_element.appendChild(child);
                m_element.appendChild(doc.createTextNode(LINE_SEPARATOR));
            }
        }
        else {
            Node n= createElement(key, value);
            m_element.appendChild(doc.createTextNode(LINE_SEPARATOR));
            m_element.appendChild(n);
            m_element.appendChild(doc.createTextNode(LINE_SEPARATOR));
        }
        return old;
    }

    Element createElement(Object key, Object value) {
        Document doc= m_element.getOwnerDocument();
        Element element= doc.createElement(key.toString());
        if (value instanceof Map) {
            for (Iterator i= ((Map)value).entrySet().iterator(); i.hasNext();) {
                element.appendChild(doc.createTextNode(LINE_SEPARATOR));
                Map.Entry entry= (Map.Entry)i.next();
                Element child= createElement(entry.getKey(), entry.getValue());
                element.appendChild(child);
            }
            element.appendChild(doc.createTextNode(LINE_SEPARATOR));
        }
        else
            element.appendChild(doc.createTextNode(value.toString()));
        return element;
    }


    public Object remove(Object key) {
        Object old= get(key);
        if (old != null) {
            NodeList nl= getNodeList();
            clearNodeList();
            Object value= null;
            for (int i= nl.getLength(); 0 < i--;) {
                Node n= nl.item(i);
                if (n.getNodeName().equals(key))
                    m_element.removeChild(n);
            }
    	}
        return old;
    }
    public void clear() {
        NodeList nl= m_element.getChildNodes();
        clearNodeList();
        Object value= null;
        for (int i= nl.getLength(); 0 < i--;)
                m_element.removeChild(nl.item(i));
    }
    public Set entrySet() {
        Set set= new java.util.AbstractSet() {
		public Iterator iterator() {
		    return new Iterator() {
                        int i= 0;
                        Collection keys= new HashSet(getNodeList().getLength());
			public boolean hasNext() {
                            NodeList l= getNodeList();
                            boolean hasNext= false;
                            for (; i < l.getLength(); i++) {
                                Node n= l.item(i);
                                if (!keys.contains(n.getNodeName()))
                                    return true;
                            }
                            return false;
			}

			public Object next() {
                            NodeList l= getNodeList();
                            for (; i < l.getLength(); i++) {
                                Node n= l.item(i);
                                final String key= n.getNodeName();
                                if (!keys.contains(key)) {
                                    keys.add(key);
                                    final Object value= get(key);
                                    Map.Entry entry= new Map.Entry() {
                                        public Object getKey() { return key; }
                                        public Object getValue() { return value; }
                                        public Object setValue(Object value) {
                                            return put(key, value);
                                        }
                                    };
                                    return entry;
                                }
                            }
                            return null;
			}

			public void remove() {
                            NodeList l= getNodeList();
                            Node n= l.item(i);
                            String key= n.getNodeName();
                            ElementMap.this.remove(key);
			}
                    };
                }

		public int size() {
		    return ElementMap.this.size();
		}

		public boolean contains(Object v) {
		    return ElementMap.this.containsValue(v);
		}
	    };
	return set;
    }

    private Collection equals_object_list= null;
    /**
     * Fixes a problem in the AbstractMap.equals method that causes the method
     * to generate a stack overflow when a Map recursively contains itself.
     */
    public synchronized boolean equals(Object o) {
    	if (o == this)
    	    return true;

    	if (!(o instanceof Map))
    	    return false;

    	Map t = (Map) o;
    	if (t.size() != size())
    	    return false;

        /*
        If, during the process of comparing all the elements in this map, we
        have recursively called into this method, and we are being asked to
        compare an object that has already been checked then just return true.
        */
        boolean clear_equals_object_list= false;
        if (equals_object_list == null) {
            clear_equals_object_list= true;
            equals_object_list= new HashSet();
        }
        if (equals_object_list.contains(o))
            return true;
        equals_object_list.add(o);

        boolean isEqual= true;
    	Iterator i = entrySet().iterator();
    	while (i.hasNext()) {
    	    Map.Entry e = (Map.Entry) i.next();
    	    Object key = e.getKey();
    	    Object value = e.getValue();

    	    if (value == null) {
        	if (!(t.get(key)==null && t.containsKey(key))) {
                    isEqual= false;
                    break;
                }
    	    } else {
        	if (!value.equals(t.get(key))) {
                    isEqual= false;
                    break;
                }
    	    }
    	}

        if (clear_equals_object_list)
            equals_object_list= null;

    	return isEqual;
    }


    public void load(InputStream in)
    throws IOException, SAXException
    {
        try {
            /*
            read XML configuration from the givin stream
            */
            DocumentBuilderFactory docBuilderFactory= DocumentBuilderFactory.newInstance();
            DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
            Document docRequest= docBuilder.parse(in);
            Element docElement= docRequest.getDocumentElement();
            docElement.normalize();

            if (m_element == null) {
                m_element= docElement;
            }
            else {
                ElementMap map= new ElementMap(docElement);
                putAll(map);
            }
        }
        catch (ParserConfigurationException x) {
            x.printStackTrace();
            throw new IOException(x.getMessage());
        }
    }

    public void save(OutputStream out)
    throws IOException
    {
        if (m_element == null)
            return;

       XMLUtils.separatize(m_element);

        String txt= m_element.toString();
        Writer writer= new PrintWriter(out);
        writer.write(txt);
        writer.flush();
    }
}
