/*
 *  EventListenerList.java
 */
package org.servicetango.util;

import java.io.*;
import java.util.*;

/**
 * A class which holds a list of EventListeners.
 */
public class EventListenerList
implements Serializable
{
    public static class Entry
    implements Serializable
    {
        Class t;
        EventListener l;
        public Entry(Class t, EventListener l) {
            this.t= t;
            this.l= l;
        }
        public Class getListenerClass() { return t; }
        public EventListener getListener() { return l; }
        public boolean equals(Object o) {
            if (o instanceof Entry) {
                Entry e= (Entry)o;
                Class c= e.getClass();
                if (!c.equals(t))
                    return false;
                EventListener el= e.getListener();
                return el.equals(l);
            }
            return super.equals(o);
        }
    };

    /* A null array to be shared by all empty listener lists*/
    private final static Entry[] NULL_ARRAY = new Entry[0];
    /* The list of ListenerType - Listener pairs */
    protected transient Entry[] listenerList = NULL_ARRAY;

    public Entry[] getListenerList() {
	return listenerList;
    }

    public synchronized EventListener[] getListenerList(Class t) {
        ArrayList list= new ArrayList(listenerList.length);
	Entry[] lList = listenerList;
	for (int i = 0; i < lList.length; i++) {
	    if (t == lList[i].getListenerClass())
                list.add(lList[i].getListener());
	}
        EventListener[] el= new EventListener[list.size()];
        list.toArray(el);
	return el;
    }

    public synchronized Entry[] getEntries(Class t) {
        ArrayList list= new ArrayList(listenerList.length);
	Entry[] lList = listenerList;
	for (int i = 0; i < lList.length; i++) {
	    if (t == lList[i].getListenerClass())
                list.add(lList[i]);
	}
        Entry[] el= new Entry[list.size()];
        list.toArray(el);
	return el;
    }

    /**
     * Return the total number of listeners for this listenerlist
     */
    public int getListenerCount() {
	return listenerList.length;
    }

    /**
     * Return the total number of listeners of the supplied type
     * for this listenerlist.
     */
    public int getListenerCount(Class t) {
	int count = 0;
	Entry[] lList = listenerList;
	for (int i = 0; i < lList.length; i++) {
	    if (t == lList[i].getListenerClass())
		count++;
	}
	return count;
    }
    public final void add(Class t, EventListener l) {
        add(new Entry(t, l));
    }
    /**
     * Add the listener as a listener of the specified type.
     * @param t the type of the listener to be added
     * @param l the listener to be added
     */
    public synchronized void add(Entry e) {
        Class t= e.getListenerClass();
        EventListener l= e.getListener();
	if (!t.isInstance(l)) {
	    throw new IllegalArgumentException("Listener " + l +
					 " is not of type " + t);
	}
	if (l ==null) {
	    throw new IllegalArgumentException("Listener " + l +
					 " is null");
	}
	if (listenerList == NULL_ARRAY) {
	    // if this is the first listener added,
	    // initialize the lists
	    listenerList = new Entry[] { e };
	} else {
	    // Otherwise copy the array and add the new listener
	    int i = listenerList.length;
	    Entry[] tmp = new Entry[i+1];
	    System.arraycopy(listenerList, 0, tmp, 0, i);
	    tmp[i] = e;
	    listenerList = tmp;
	}
    }

    public void remove(Entry e) {
        remove(e.getListenerClass(), e.getListener());
    }
    /**
     * Remove the listener as a listener of the specified type.
     * @param t the type of the listener to be removed
     * @param l the listener to be removed
     */
    public synchronized void remove(Class t, EventListener l) {
	if (!t.isInstance(l)) {
	    throw new IllegalArgumentException("Listener " + l +
					 " is not of type " + t);
	}
	if (l ==null) {
	    throw new IllegalArgumentException("Listener " + l +
					 " is null");
	}

	// Is l on the list?
	int index = -1;
	for (int i = 0; i < listenerList.length; i++) {
            Entry e= listenerList[i];
	    if ((e.getListenerClass()==t) && (e.getListener().equals(l) == true)) {
		index = i;
		break;
	    }
	}

	// If so,  remove it
	if (index != -1) {
	    Entry[] tmp = new Entry[listenerList.length-1];
	    // Copy the list up to index
	    System.arraycopy(listenerList, 0, tmp, 0, index);
	    // Copy from two past the index, up to
	    // the end of tmp (which is two elements
	    // shorter than the old list)
	    if (index < tmp.length)
		System.arraycopy(listenerList, index+1, tmp, index,
				 tmp.length - index);
	    // set the listener array to the new array or null
	    listenerList = (tmp.length == 0) ? NULL_ARRAY : tmp;
	    }
    }

    // Serialization support.
    private void writeObject(ObjectOutputStream s) throws IOException {
	Entry[] lList = listenerList;
	s.defaultWriteObject();

	// Save the non-null event listeners:
	for (int i = 0; i < lList.length; i++) {
            Entry e= lList[i];
	    Class t = e.getListenerClass();
	    EventListener l = e.getListener();
	    if ((l!=null) && (l instanceof Serializable)) {
		s.writeObject(t.getName());
		s.writeObject(l);
	    }
	}

	s.writeObject(null);
    }

    private void readObject(ObjectInputStream s)
	throws IOException, ClassNotFoundException {
        listenerList = NULL_ARRAY;
	s.defaultReadObject();
	Object listenerTypeOrNull;

	while (null != (listenerTypeOrNull = s.readObject())) {
	    EventListener l = (EventListener)s.readObject();
	    add(Class.forName((String)listenerTypeOrNull), l);
	}
    }

    /**
     * Return a string representation of the EventListenerList.
     */
    public String toString() {
	Entry[] lList = listenerList;
	String s = "EventListenerList: ";
	s += lList.length + " listeners: ";
	for (int i = 0 ; i <= lList.length ; i++) {
            Entry e= lList[i];
	    s += " type " + e.getListenerClass().getName();
	    s += " listener " + e.getListener();
	}
	return s;
    }
}
