/*
   LangUtils.java

   Copyright (c) 2000 by ted stockwell
   All Rights Reserved
*/

package org.servicetango.util;

import java.util.Date;
import java.util.TimeZone;
import java.util.SimpleTimeZone;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.text.DateFormat;

/**
 * Miscellaneouse Collection utilites
 *
 * @author ted stockwell [emorning@mediaone.net]
 */
public class LangUtils {

    public static final int TYPE_UNKNOWN= 0;
    public static final int TYPE_STRING= 1;
    public static final int TYPE_INTEGER= 2;
    public static final int TYPE_BYTE= 3;
    public static final int TYPE_BOOLEAN= 4;
    public static final int TYPE_LONG= 5;
    public static final int TYPE_CHARACTER= 6;
    public static final int TYPE_DOUBLE= 7;
    public static final int TYPE_FLOAT= 8;
    public static final int TYPE_SHORT= 9;
    public static final int TYPE_BIG_DECIMAL= 10;
    public static final int TYPE_BIG_INTEGER= 11;
    public static final int TYPE_DATE= 12;
    public static final int TYPE_TIME_ZONE= 13;


    /**
    *   Returns a code that indicates the type of the given object.
    *   The type code is one of the following:
            TYPE_STRING : An instance of java.lang.String
            TYPE_INTEGER : An instance of java.lang.Integer
            TYPE_BYTE : An instance of java.lang.Byte
            TYPE_BOOLEAN : An instance of java.lang.Boolean
            TYPE_LONG : An instance of java.lang.Long
            TYPE_CHARACTER : An instance of java.lang.Character
            TYPE_DOUBLE : An instance of java.lang.Double
            TYPE_FLOAT : An instance of java.lang.Float
            TYPE_SHORT : An instance of java.lang.Short
            TYPE_BIG_DECIMAL : An instance of java.math.BigDecimal
            TYPE_BIG_INTEGER : An instance of java.math.BigInteger
            TYPE_DATE : An instance of java.lang.Date
            TYPE_TIME_ZONE : An instance of java.lang.TimeZone

            TYPE_UNKNOWN : Anything that's not one of the above
    */
    final static public int getType(Object obj) {
        if (obj instanceof String)
            return TYPE_STRING;
        if (obj instanceof Number) {
            if (obj instanceof Integer)
                return TYPE_INTEGER;
            if (obj instanceof Byte)
                return TYPE_BYTE;
            if (obj instanceof Long)
                return TYPE_LONG;
            if (obj instanceof Double)
                return TYPE_DOUBLE;
            if (obj instanceof Float)
                return TYPE_FLOAT;
            if (obj instanceof Short)
                return TYPE_SHORT;
            return TYPE_UNKNOWN;
        }
        if (obj instanceof Boolean)
            return TYPE_BOOLEAN;
        if (obj instanceof Character)
            return TYPE_CHARACTER;
        if (obj instanceof java.math.BigDecimal)
            return TYPE_BIG_DECIMAL;
        if (obj instanceof java.math.BigInteger)
            return TYPE_BIG_INTEGER;
        if (obj instanceof java.util.Date)
            return TYPE_DATE;
        if (obj instanceof java.util.TimeZone)
            return TYPE_TIME_ZONE;
        return TYPE_UNKNOWN;
    }

    /**
    *   Convenience method that returns a value as a date
    *   Returns null if the given value cannot be converted to a Date.
    */
    static public Date toDate(Object o) {
        if (o == null)
            return null;
        if (o instanceof Date)
            return (Date)o;
        Date result= null;

        // check to see if the object has a toDate() method...
        try {
            Method toDateMethod= o.getClass().getMethod("toDate", new Class[0]);
            return (Date)toDateMethod.invoke(o, new Object[0]);
        }
        catch (Throwable t) {
        }

        // try converting it to a String and parsing the String
        String value= o.toString();
        try {
            result= DATE_FORMAT.parse(value);
        }
        catch (Throwable t) {
            long longValue= toLong(o, (long)-1);
            if (0 <= longValue)
                return new Date(longValue);
        }

        return result;
    }
    static final DateFormat DATE_FORMAT= DateFormat.getDateTimeInstance();

    /**
    *   Convenience method that returns a value as a time zone
    *   Returns null if the given value cannot be converted to a TimeZone.
    */
    static public TimeZone toTimeZone(Object o) {
        if (o == null)
            return null;
        if (o instanceof TimeZone)
            return (TimeZone)o;
        TimeZone result= null;

        if (o instanceof Integer) {
            int rawOffset= ((Integer)o).intValue();
            String ids[]= TimeZone.getAvailableIDs(rawOffset);
            String id= (ids != null && 0 < ids.length) ? ids[0] : "";
            o= new SimpleTimeZone(rawOffset, id);
        }

        return result;
    }

    /**
    *   Convenience method that returns a value as an Integer
    */
	static public Integer toInteger(Object obj) {
        if (obj == null)
            return null;
        if (obj instanceof Number)
            return new Integer(((Number)obj).intValue());
        try {
            return Integer.valueOf(obj.toString());
        }
        catch (Throwable t) {
        }
        return null;
	}

    /**
    *   Convenience method that returns a value as an int
    */
	static public int toInt(Object obj, int defaultValue)	{
        if (obj instanceof Number)
            return ((Number)obj).intValue();
        try {
            return Integer.parseInt(obj.toString());
        }
        catch (Throwable t) {
        }
        return defaultValue;
	}


    /**
    *   Convenience method that returns a value as a long
    */
	static public long toLong(Object obj, long defaultValue)	{
        if (obj instanceof Number)
            return ((Number)obj).longValue();
        try {
            return Long.parseLong(obj.toString());
        }
        catch (Throwable t) {
        }
        return defaultValue;
	}

    /**
    *   Convenience method that returns a value as a byte
    */
	static public byte toByte(Object obj, byte defaultValue)	{
        if (obj instanceof Number)
            return ((Number)obj).byteValue();
        try {
            return Byte.parseByte(obj.toString());
        }
        catch (Throwable t) {
        }
        return defaultValue;
	}

    /**
    *   Convenience method that returns a value as a boolean
    */
	static public boolean toBoolean(Object obj, boolean defaultValue)	{
        int result;
        if (obj == null)
            return defaultValue;
        if (obj instanceof Boolean) {
            return ((Boolean)obj).booleanValue();
        } else if (obj instanceof Number) {
            return (((Number)obj).intValue() == 0) ? false : true;
        } else if (obj instanceof String) {
            String s= (String)obj;
            if (s.equalsIgnoreCase("true")) {
                return true;
            }
            else if (s.equalsIgnoreCase("false")) {
                return false;
            }
            else {
                try {
                    return Integer.parseInt((String)obj) != 0;
                }
                catch (Throwable t) {
                    return false;
                }
            }
        }
        return defaultValue;
	}
    /**
    *   Convenience method that returns a value as a boolean
    */
	static public Boolean toBoolean(Object obj)	{
        Boolean result= null;
        if (obj == null)
            return null;
        if (obj instanceof Boolean) {
            return (Boolean)obj;
        } else if (obj instanceof Number) {
            return (((Number)obj).intValue() == 0) ? Boolean.FALSE : Boolean.TRUE;
        } else if (obj instanceof String) {
            String s= (String)obj;
            if (s.equalsIgnoreCase("true")) {
                return Boolean.TRUE;
            }
            else if (s.equalsIgnoreCase("false")) {
                return Boolean.FALSE;
            }
            else {
                try {
                    return new Boolean(Integer.parseInt((String)obj) != 0);
                }
                catch (Throwable t) {
                    return Boolean.FALSE;
                }
            }
        }
        return null;
	}

    static public final String toString(Object obj) {
        if (obj == null)
            return null;
        return obj.toString();
    }
}

