/*
   ProfileMap.java

   Copyright (c) 2000 by ted stockwell
   All Rights Reserved
*/
package org.servicetango.util;
import java.io.*;
import java.text.*;
import java.sql.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Date;
import java.util.*;

/**
 * Adds methods to load and save a Map.
 * Only Strings should be used as keys to a ProfileMap.
 * Values may be any class but Strings are preferred.
 * Just like java.util.Properties except that this class preserves comments
 * associated with properties when the file is read and written.
 * More like a Windows profile.
 *
 * @author  ted stockwell [emorning@mediaone.net]
 */
public class ProfileMap extends FilterMap
{
    public final String LINE_SEPARATOR= System.getProperty("line.separator");


    protected HashMap m_comments= new HashMap();
    protected String[] m_headers= new String[0];

    /**
     * used to preserve the order of the original file
     */
    protected List m_list= new ArrayList();

    public ProfileMap(Map map) {
	    super(map);
    }
    public ProfileMap(int count) {
	    super(new HashMap(count));
    }
    public ProfileMap() {
	    super(new HashMap());
    }

    /**

    Returns a reference to this ProfileMap's underlying Map

    */

    protected Map getMap() {

        return super.getMap();

    }

    /**
    Change the underlying Map used by this ProfileMap

    */

    protected void setMap(Map value) {

        super.setMap(value);

    }


    public String getProperty(String key) {
        return (String)get(key);
    }
    public String getProperty(String key, String defaultValue) {
        Object o= get(key);
        return (o == null) ? defaultValue : (String)o;
    }

    /**
     * Reads a property list from an input stream.
     *
     * @param      in   the input stream.
     * @exception  IOException  if an error occurred when reading from the
     *               input stream.
     */
    public synchronized void load(InputStream inputStream) throws IOException
    {
        int ch;

        // read in the entire string
        Reader inputStreamReader=
            new BufferedReader(
                new InputStreamReader(
                    inputStream));
        StringBuffer stringBuffer= new StringBuffer();
        for (;(ch= inputStreamReader.read()) != -1;)
            stringBuffer.append((char)ch);
        String string= stringBuffer.toString();
        if (string.length() <= 0)
            return;
        if (!string.endsWith(LINE_SEPARATOR))
            string+= LINE_SEPARATOR;

        StringReader in= new StringReader(string);
        Vector comments= new Vector();
        Vector headers= new Vector();
        boolean isHeader= false;

	    ch = in.read();
	    while (ch != -1)
        {
	        switch (ch) {
	          case '!':
                isHeader= true;
                // fall through

	          case '#':
              {
                stringBuffer.setLength(0);
		        for (;;)
                {
		           ch = in.read();
		            if ((ch < 0) || (ch == '\n') || (ch == '\r'))
                        break;
                    stringBuffer.append((char)ch);
		        }
                String comment= stringBuffer.toString();
                if (isHeader)
                {
                    headers.addElement(comment);
                    isHeader= false;
                }
                else
                    comments.addElement(comment);
              }
		    continue;

	          case '\n':
	          case '\r':
	          case ' ':
	          case '\t':
		    ch = in.read();
		    continue;
	        }

	        // Read the key
	        StringBuffer key = new StringBuffer();
	        while ((ch >= 0) && (ch != '=') && (ch != ':') &&
		       (ch != ' ') && (ch != '\t') && (ch != '\n') && (ch != '\r')) {
		    key.append((char)ch);
		    ch = in.read();
	        }
	        while ((ch == ' ') || (ch == '\t')) {
		    ch = in.read();
	        }
	        if ((ch == '=') || (ch == ':')) {
		    ch = in.read();
	        }
	        while ((ch == ' ') || (ch == '\t')) {
		    ch = in.read();
	        }

	        // Read the value
	        StringBuffer val = new StringBuffer();
	        while ((ch >= 0) && (ch != '\n') && (ch != '\r')) {
		    int next = 0;
		    if (ch == '\\') {
		        switch (ch = in.read()) {
		          case '\r':
			    if (((ch = in.read()) == '\n') ||
			        (ch == ' ') || (ch == '\t')) {
			      // fall thru to '\n' case
			    } else continue;
		          case '\n':
			    while (((ch = in.read()) == ' ') || (ch == '\t'));
			    continue;
		          case 't': ch = '\t'; next = in.read(); break;
		          case 'n': ch = '\n'; next = in.read(); break;
		          case 'r': ch = '\r'; next = in.read(); break;
		          case 'u': {
			    while ((ch = in.read()) == 'u');
			    int d = 0;
		          loop:
			    for (int i = 0 ; i < 4 ; i++) {
			        next = in.read();
			        switch (ch) {
			          case '0': case '1': case '2': case '3': case '4':
			          case '5': case '6': case '7': case '8': case '9':
				    d = (d << 4) + ch - '0';
				    break;
			          case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
				    d = (d << 4) + 10 + ch - 'a';
				    break;
			          case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
				    d = (d << 4) + 10 + ch - 'A';
				    break;
			          default:
				    break loop;
			        }
			        ch = next;
			    }
			    ch = d;
			    break;
		          }
		          default: next = in.read(); break;
		        }
		    } else {
		        next = in.read();
		    }
		    val.append((char)ch);
		    ch = next;
	        }

	        //System.out.println(key + " = '" + val + "'");
            String keyString= key.toString();
	        put(keyString, val.toString());

            int commentCount= comments.size();
            if (0 < commentCount)
            {
                String[] lines= new String[commentCount];
                comments.copyInto(lines);
                if (m_comments == null)
                    m_comments= new HashMap();
                m_comments.put(keyString, lines);
                comments.removeAllElements();
            }
	    }

        int hdrCount= headers.size();
        if (0 < hdrCount)
        {
            m_headers= new String[hdrCount];
            for (int i= headers.size(); 0 < i--;)
                m_headers[i]= (String)headers.elementAt(i);
        }
    }

    public Object put(Object key, Object value)    {
        Object o= super.put(key, value);
        if (!m_list.contains(key))
            m_list.add(key);
        return o;
    }

    public void putAll(Map t) {
        for (Iterator i= t.entrySet().iterator(); i.hasNext();) {
            Map.Entry entry= (Map.Entry)i.next();
            put(entry.getKey(), entry.getValue());
        }
    }

    public Object remove(Object key) {
        Object o= super.remove(key);
        m_list.remove(key);
        return o;
    }

    /**
     * Stores this property list to the specified output stream. The
     * string header is printed as a comment at the beginning of the stream.
     *
     * @param   out      an output stream.
     * @param   header   a description of the property list.
     */
    public synchronized void save(OutputStream out, String header) {
        OutputStreamWriter outputStreamWriter= new OutputStreamWriter(out);
        PrintWriter prnt= new PrintWriter(outputStreamWriter);
        boolean localize = true;

        //OutputStream localOut = Runtime.getRuntime().getLocalizedOutputStream(out);
        //PrintStream prnt = new PrintStream(localOut, false);
        //boolean localize = localOut != out;

        if (header != null) {
            prnt.write('!');
            prnt.println(header);
        }

        for (int iHeader= 0; iHeader < m_headers.length; iHeader++) {
            prnt.write('!');
            prnt.print(m_headers[iHeader]);
            prnt.println();
        }

        boolean lastCommented= false;
        String lastKey= "";
        for (Iterator e = m_list.iterator() ; e.hasNext() ;) {
            String key = (String)e.next();

            String[] comments= (String[])m_comments.get(key);
            boolean commented= false;
            if (comments != null) {
                prnt.println();
                for (int iComment= 0; iComment < comments.length; iComment++) {
                    prnt.write('#');
                    prnt.println(comments[iComment]);
                }
                commented= true;
            }

            /*
            If the last property had a comment and this property is not
            commented and the two properties do not have a common
            prefix then insert a blank line
            */
            if (commented == false  &&  lastCommented == true) {
                int i;
                String lastPrefix= lastKey;
                String prefix= key;
                if (0 < (i= lastKey.indexOf('.')))
                    lastPrefix= lastKey.substring(0, i);
                if (0 < (i= key.indexOf('.')))
                    prefix= key.substring(0, i);
                if (!lastPrefix.equals(prefix))
                    prnt.println();
            }

            prnt.print(key);
            prnt.write('=');

            String val = LangUtils.toString(get(key));
            int len = val.length();
            boolean empty = false;

            for (int i = 0 ; i < len ; i++) {
                int ch = val.charAt(i);

                switch (ch) {
                case '\\': prnt.write('\\'); prnt.write('\\'); break;
                case '\t': prnt.write('\\'); prnt.write('t'); break;
                case '\n': prnt.write('\\'); prnt.write('n'); break;
                case '\r': prnt.write('\\'); prnt.write('r'); break;

                default:
                    if ((ch < ' ') || (ch >= 127) || (empty && (ch == ' '))) {
                        if ((ch > 255) && localize) {
                            prnt.write(ch);
                        } else {
                            prnt.write('\\');
                            prnt.write('u');
                            prnt.write(toHex((ch >> 12) & 0xF));
                            prnt.write(toHex((ch >>  8) & 0xF));
                            prnt.write(toHex((ch >>  4) & 0xF));
                            prnt.write(toHex((ch >>  0) & 0xF));
                        }
                    } else {
                        prnt.write(ch);
                    }
                }
                empty = false;
            }
            prnt.write('\n');

            lastCommented= commented;
            lastKey= key;
        }
        prnt.flush();
    }

    /**
     * Convert a nibble to a hex character
     * @param	nibble	the nibble to convert.
     */
    private static char toHex(int nibble) {
	return hexDigit[(nibble & 0xF)];
    }
    /** A table of hex digits */
    private static char[] hexDigit = {
	'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
    };

}
