/*
   XMLUtils.java

   Copyright (C) 1999 by GolfTango
   All Rights Reserved
*/

package org.servicetango.util;
import org.w3c.dom.*;
import java.util.*;
import java.text.*;



/**
 * Various XML utilities that make it easier to use the classes in the
 * javax.xml.parsers, org.w3c.dom, and org.xml.sax packages.
 *
 * @author  ted stockwell [emorning@mediaone.net]
 */
public class XMLUtils
{
    static final String LINE_SEPARATOR= (String)System.getProperty("line.separator");

    static public class Exception extends java.lang.Exception {
        Exception(String msg) {
            super(msg);
        }
    }


    /**
     * Converts the given element to Map.  The entire element is recursively
     * processed to create a tree of Map and List objects.
     *
     * This method assumes the following pattern is true;
     * The given element is the "document" element, it is not included in the
     * returned Map object.  Tag names are converted into property names in the
     * resulting Map.  Tag contents (the contents between the beginning and
     * ending tags) are conerted to property values.  For any element, if there
     * is more than a single instance of the Element with the same name then a
     * List object is created and the List is populated with Map objects that
     * represent the instances.
     */
    static public Map toMap(Element docElement)
    throws Exception
    {
        Object m= _toValue(docElement);
        if (m == null  ||  !(m instanceof Map))
            m= new HashMap(0);
        return (Map)m;
    }
    static Object _toValue(Element docElement)
    throws Exception
    {
        NodeList children= getElementChildNodes(docElement);
        int count;
        if (children == null  ||  (count= children.getLength()) <= 0)
            return docElement.getNodeValue();
        Map m= new HashMap(count);
        for (int i= 0; i < count; i++) {
            Element n= (Element)children.item(i);
            String key= n.getNodeName();
            Object value= _toValue(n);
            m.put(key, value);
        }
        return m;
    }

    final static public NodeList getElementChildNodes(Node node) {
        return getChildNodes(node, Node.ELEMENT_NODE);
    }
    final static public NodeList getTextChildNodes(Node node) {
        return getChildNodes(node, Node.TEXT_NODE);
    }
    static public NodeList getChildNodes(Node node, short nodeType) {
        if (node == null)
            return new NodeList() {
                public int getLength() { return 0; }
                public Node item(int i) { return null; }
            };
        NodeList children= node.getChildNodes();
        int count= (children == null) ? 0 : children.getLength();
        final List l= new ArrayList(count);
        for (int i= 0; i < count; i++) {
            Node n= children.item(i);
            if (n.getNodeType() == nodeType)
                l.add(n);
        }
        return new NodeList() {
          public int getLength() { return l.size(); }
          public Node item(int i) { return (Node)l.get(i); }
        };
    }

    static public String toDateString(Date date) {
        return DateUtils.toISO8601String(date);
    }

    static public String getValueByTagName(Element docElement, String tagName)
    throws Exception
    {
        NodeList nodelistID= docElement.getElementsByTagName(tagName);
        if (nodelistID.getLength() <= 0)
            throw new Exception("Missing "+tagName+" tag");
        Element elementID= (Element)nodelistID.item(0);
        NodeList childnodesID= elementID.getChildNodes();
        if (childnodesID.getLength() <= 0)
            throw new Exception(tagName+" element is empty");
        String value= childnodesID.item(0).getNodeValue();
        return value.trim();
    }

    /**
     * @return  A list of Element objects
     */
    static public NodeList getAllElementsByTagName(Element docElement, String tagName)
    throws Exception
    {
        NodeList nodelistID= docElement.getElementsByTagName(tagName);
        if (nodelistID.getLength() <= 0)
            throw new Exception("Missing "+tagName+" tag");
        return nodelistID;
    }

    static public String toBooleanString(boolean value) {
        return value ? "Yes" : "No";
    }


    /**
     * Like Element.normalize except that all empty text elements are replaced
     * by line separators.
     * This method is normally used before writing a document that is meant for
     * human consumption (the line separators make the document readable).
     */
    static public void separatize(Element element) {
        element.normalize();
        NodeList nl= element.getChildNodes();
        for (int i= nl.getLength(); 0 < i--;) {
            Node n= nl.item(i);
            int typ= n.getNodeType();
            if (typ == Node.TEXT_NODE) {
                String txt= n.getNodeValue();
                if (n.getNodeValue().equals(" "))
                    n.setNodeValue(LINE_SEPARATOR+LINE_SEPARATOR);
            }
            else if (typ == Node.ELEMENT_NODE) {
                _separatize((Element)n);
            }
        }
    }
    static private void _separatize(Element element) {
        NodeList nl= element.getChildNodes();
        for (int i= nl.getLength(); 0 < i--;) {
            Node n= nl.item(i);
            int typ= n.getNodeType();
            if (typ == Node.TEXT_NODE) {
                String txt= n.getNodeValue();
                if (n.getNodeValue().equals(" "))
                    n.setNodeValue(LINE_SEPARATOR);
            }
            else if (typ == Node.ELEMENT_NODE) {
                _separatize((Element)n);
            }
        }
    }
}

